from .atmospheric_model import AtmosphericLayer, power_spectral_density_von_karman, fried_parameter_from_Cn_squared
import numpy as np
from ..util import SpectralNoiseFactoryMultiscale

import copy

class FiniteAtmosphericLayer(AtmosphericLayer):
    '''An atmospheric layer simulating atmospheric turbulence.

    This atmospheric layer is finite. This means that it will wrap
    after translation by more than `oversampling` times the extent of
    the input grid.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the incoming wavefront is defined.
    Cn_squared : scalar
        The integrated strength of the turbulence for this layer.
    L0 : scalar
        The outer scale for the atmospheric turbulence of this layer.
        The default is infinity.
    velocity : scalar or array_like
        The wind speed for this atmospheric layer. If this is a scalar,
        the wind will be along x. If this is a 2D array, then the values
        are interpreted as the wind speed along x and y. The default is
        zero.
    height : scalar
        The height of the atmospheric layer. By itself, this value has no
        influence, but it'll be used by the AtmosphericModel to perform
        inter-layer propagations.
    oversampling : scalar
        The amount of oversampling in the Fourier space. The atmospheric layer
        will wrap after translation by more than `oversampling` times the extent of
        the input grid. The default is 2.
    seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
        A seed to initialize the spectral noise. If None, then fresh, unpredictable
        entry will be pulled from the OS. If an int or array of ints, then it will
        be passed to a numpy.SeedSequency to derive the initial BitGenerator state.
        If a BitGenerator or Generator are passed, these will be wrapped and used
        instead. Default: None.
    '''
    def __init__(self, input_grid, Cn_squared=None, L0=np.inf, velocity=0, height=0, oversampling=2, seed=None):
        self._noise = None
        self._achromatic_screen = None

        AtmosphericLayer.__init__(self, input_grid, Cn_squared, L0, velocity, height)

        self.oversampling = oversampling
        self.center = np.zeros(2)

        self._original_rng = np.random.default_rng(seed)

        self.reset()

    def reset(self, make_independent_realization=False):
        '''Reset the atmospheric layer to t=0.

        Parameters
        ----------
        make_independent_realization : boolean
            Whether to start an independent realization of the noise for the
            atmospheric layer or not. When this is False, the exact same phase
            screens will be generated as in the first run, as long as the Cn^2
            and outer scale are the same. This allows for testing of multiple
            runs of a control system with different control parameters. When
            this is True, an independent realization of the atmospheric layer
            will be generated. This is useful for Monte-Carlo-style computations.
            The default is False.
        '''
        if make_independent_realization:
            # Reset the original random generator to the current one. This
            # will essentially reset the randomness.
            self._original_rng = copy.copy(self.rng)
        else:
            # Make a copy of the original random generator. This copy will be
            # used as the source for all randomness.
            self.rng = copy.copy(self._original_rng)

        self.psd = power_spectral_density_von_karman(fried_parameter_from_Cn_squared(self.Cn_squared, 1), self.L0)

        self.noise_factory = SpectralNoiseFactoryMultiscale(self.psd, self.input_grid, self.oversampling)
        self._noise = self.noise_factory.make_random(self.rng)

        self._achromatic_screen = None

    @property
    def noise(self):
        '''The (unshifted) spectral noise of this layer.

        This property is not intended to be used by the user.
        '''
        if self._noise is None:
            self.reset()

        return self._noise

    @property
    def achromatic_screen(self):
        '''The phase of this layer for a wavelength of one.

        This property is not intended to be used by the user.
        '''
        if self._achromatic_screen is None:
            self._achromatic_screen = self.noise.shifted(self.center)()

        return self._achromatic_screen

    def phase_for(self, wavelength):
        '''Compute the phase at a certain wavelength.

        Parameters
        ----------
        wavelength : scalar
            The wavelength of the light for which to compute the phase screen.

        Returns
        -------
        Field
            The computed phase screen.
        '''
        return self.achromatic_screen / wavelength

    def evolve_until(self, t):
        '''Evolve the atmospheric layer until a certain time.

        Parameters
        ----------
        t : scalar
            The new time to evolve the phase screen to.
        '''
        self.center = self.velocity * t
        self._achromatic_screen = None

    @property
    def Cn_squared(self):  # noqa: N802
        '''The integrated strength of the turbulence for this layer.
        '''
        return self._Cn_squared

    @Cn_squared.setter
    def Cn_squared(self, Cn_squared):  # noqa: N802
        self._Cn_squared = Cn_squared
        self._noise = None

    @property
    def outer_scale(self):
        '''The outer scale of the turbulence for this layer.
        '''
        return self._L0

    @outer_scale.setter
    def L0(self, L0):  # noqa: N802
        self._L0 = L0
        self._noise = None



from __future__ import division

import numpy as np
from .fourier_transform import FourierTransform, multiplex_for_tensor_fields, _get_float_and_complex_dtype
from ..field import Field, CartesianGrid, RegularCoords
from ..config import Configuration
import numexpr as ne

from .._math import fft as _fft_module

def make_fft_grid(input_grid, q=1, fov=1, shift=0):
    '''Calculate the grid returned by a Fast Fourier Transform.

    Parameters
    ----------
    input_grid : Grid
        The grid defining the sampling in the real domain..
    q : scalar or array_like
        The amount of zeropadding to perform. A value of 1 denotes no zeropadding.
    fov : scalar or array_like
        The amount of cropping to perform in the Fourier domain.
    shift : scalar or array_like
        The shift to apply on the output grid.

    Returns
    -------
    Grid
        The grid defining the sampling in the Fourier domain.
    '''
    q = np.ones(input_grid.ndim, dtype='float') * q
    fov = np.ones(input_grid.ndim, dtype='float') * fov
    shift = np.ones(input_grid.ndim, dtype='float') * shift

    # Correct q for a discrete zero padding of the input grid.
    q = np.round(q * input_grid.dims) / input_grid.dims

    # Check assumptions
    if not input_grid.is_regular:
        raise ValueError('The input_grid must be regular.')
    if not input_grid.is_('cartesian'):
        raise ValueError('The input_grid must be cartesian.')

    delta = (2 * np.pi / (input_grid.delta * input_grid.dims)) / q
    dims = (input_grid.dims * fov * q).astype('int')
    zero = delta * (-dims / 2 + np.mod(dims, 2) * 0.5) + shift

    return CartesianGrid(RegularCoords(delta, dims, zero))

def get_fft_parameters(fft_grid, input_grid):
    '''Try to reconstruct the FFT parameters of a grid.

    .. note::
        Not every grid is an FFT grid. This function will raise a
        ValueError if this is the case. You can alternatively use
        `is_fft_grid()` to check if a grid is an FFT grid or not.

    .. note::
        The parameters that this function outputs might not
        correspond perfectly to the original FFT parameters you used.
        However, it guarantees that an FFT grid generated with these
        reconstructed parameters will create the same FFT grid as an
        FFT generated with the original parameters.

    Parameters
    ----------
    fft_grid : Grid
        A grid that corresponds to a native FFT grid of `input_grid`.
    input_grid : Grid
        The grid defining the sampling in the real domain.

    Returns
    -------
    q : ndarray
        The amount of zeropadding detected in the real domain.
    fov : ndarray
        The amount of cropping detected in the Fourier domain.
    shift : ndarray
        The amount of shifting detected in the Fourier domain.

    Raises
    ------
    ValueError
        If `fft_grid` does not correspond to an FFT grid of `input_grid`.
    '''
    if not input_grid.is_regular:
        raise ValueError('The input grid must be regular to reconstruct an fft grid.')
    if not fft_grid.is_regular:
        raise ValueError('The fft grid is not regular and therefore cannot be an fft grid.')

    q = (2 * np.pi / (input_grid.delta * input_grid.dims)) / fft_grid.delta

    if np.any(q < 1):
        raise ValueError(f'fft_grid is not an FFT grid of input_grid: q of {q} would be < 1.')

    # Check that the calculated q corresponds to an integer zeropadding.
    zeropadded_dims = q * input_grid.dims
    if np.any(np.abs(zeropadded_dims - np.round(zeropadded_dims)) > 1e-10):
        raise ValueError(f'fft_grid is not an FFT grid of input_grid: q of {q} does not correspond to an integer zeropadding.')

    # Compute fov.
    fov = fft_grid.dims / zeropadded_dims

    # Check if fov would be < 1.
    if np.any(fft_grid.dims > (zeropadded_dims + 0.5).astype('int')):
        raise ValueError(f'fft_grid is not an FFT grid of input_grid: fov of {fov} would be > 1 .')

    # Correct fov for rounding errors (floating point errors would lead to a different dims).
    dummy_fft_grid = make_fft_grid(input_grid, q, fov)
    wrong_dims = fft_grid.dims != dummy_fft_grid.dims
    fov[wrong_dims] = ((fft_grid.dims + 0.5) / (input_grid.dims * q))[wrong_dims]

    shift = fft_grid.zero - fft_grid.delta * (-fft_grid.dims / 2 + np.mod(fft_grid.dims, 2) * 0.5)
    return q, fov, shift

def is_fft_grid(grid, input_grid):
    '''Returns whether `grid` is a native FFT grid of `input_grid` or not.

    .. note::
        The function get_fft_parameters() can be used

    Parameters
    ----------
    grid : Grid
        The grid in the Fourier domain. This grid is checked.
    input_grid : Grid
        The grid in the real domain of the FFT.

    Returns
    -------
    boolean
        Whether `grid` is a native FFT grid of `input_grid` for some q, fov and shift.
    '''
    try:
        get_fft_parameters(grid, input_grid)
    except ValueError:
        return False
    return True

def _numexpr_grid_shift(shift, grid, out=None):
    '''Fast evaluation of np.exp(1j * np.dot(shift, grid.coords)) using NumExpr.

    Parameters
    ----------
    shift : array_like
        The coordinates of the shift.
    grid : Grid
        The grid on which to calculate the shift.
    out : array_like
        An existing array where the outcome is going to be stored. This must
        have the correct shape and dtype. No checking will be performed. If
        this is None, a new array will be allocated and returned.

    Returns
    -------
    array_like
        The calculated complex shift array.
    '''
    variables = {}
    command = []
    coords = grid.coords

    for i in range(grid.ndim):
        variables[f'a{i}'] = shift[i]
        variables[f'b{i}'] = coords[i]

        command.append(f'a{i} * b{i}')

    command = 'exp(1j * (' + '+'.join(command) + '))'
    return ne.evaluate(command, local_dict=variables, out=out)

class FastFourierTransform(FourierTransform):
    '''A Fast Fourier Transform (FFT) object.

    This Fourier transform calculates FFTs with zeropadding and cropping. This
    Fourier transform requires the input grid to be regular in Cartesian coordinates. Every
    number of dimensions is allowed.

    Parameters
    ----------
    input_grid : Grid
        The grid that is expected for the input field.
    q : scalar or array_like
        The amount of zeropadding to perform. A value of 1 denotes no zeropadding. A value of
        2 indicates zeropadding to twice the dimensions of the input grid. Note: as
        zeropadding has to be done by an integer number of pixels, the q will be rounded to
        the closest possible number to satisfy this constraint.
    fov : scalar or array_like
        The amount of cropping to perform in the Fourier domain. A value of 1 indicates that
        no cropping will be performed.
    shift : array_like or scalar
        The amount by which to shift the output grid. If this is a scalar, the same shift will
        be used for all dimensions.
    emulate_fftshifts : boolean or None
        Whether to emulate FFTshifts normally used in the FFT by multiplications in the
        opposite domain. Enabling this increases performance by 3x, but degrades accuracy of
        the FFT by 10x. If this is None, the choice will be determined by the configuration
        file.

    Raises
    ------
    ValueError
        If the input grid is not regular or Cartesian.
    ValueError
        If q < 1 or fov < 0 or fov > 1, both of which are impossible for an FFT to calculate.
    '''
    def __init__(self, input_grid, q=1, fov=1, shift=0, emulate_fftshifts=None):
        # Check assumptions
        if not input_grid.is_regular:
            raise ValueError('The input_grid must be regular.')
        if not input_grid.is_('cartesian'):
            raise ValueError('The input_grid must be Cartesian.')

        if np.any(q < 1):
            raise ValueError('The amount of zeropadding (q) must be larger than 1.')

        if np.any(fov < 0):
            raise ValueError('The amount of cropping (fov) must be positive.')

        self.input_grid = input_grid

        self.shape_in = input_grid.shape
        self.weights = input_grid.weights
        self.size = input_grid.size
        self.ndim = input_grid.ndim

        # Get the value from the configuration file if left at the default.
        if emulate_fftshifts is None:
            emulate_fftshifts = Configuration().fourier.fft.emulate_fftshifts
        self.emulate_fftshifts = emulate_fftshifts

        self.output_grid = make_fft_grid(input_grid, q, fov, shift)
        self.internal_grid = make_fft_grid(input_grid, q, 1)

        if np.any(self.output_grid.dims > self.internal_grid.dims):
            raise ValueError('The amount of cropping (fov) must be smaller than 1.')

        self.shape_out = self.output_grid.shape
        self.internal_shape = self.internal_grid.shape
        self.internal_array = np.zeros(self.internal_shape, 'complex')

        # Calculate the part of the array in which to insert the input field (for zeropadding).
        if np.allclose(self.internal_shape, self.shape_in):
            self.cutout_input = None
        else:
            cutout_start = (self.internal_shape / 2.).astype('int') - (self.shape_in / 2.).astype('int')
            cutout_end = cutout_start + self.shape_in
            self.cutout_input = tuple([slice(start, end) for start, end in zip(cutout_start, cutout_end)])

        # Calculate the part of the array to extract the output field (for cropping).
        if np.allclose(self.internal_shape, self.shape_out):
            self.cutout_output = None
        else:
            cutout_start = (self.internal_shape / 2.).astype('int') - (self.shape_out / 2.).astype('int')
            cutout_end = cutout_start + self.shape_out
            self.cutout_output = tuple([slice(start, end) for start, end in zip(cutout_start, cutout_end)])

        # Calculate the shift array when the input grid was shifted compared to the native shift
        # expected by the numpy FFT implementation.
        center = input_grid.zero + input_grid.delta * (np.array(input_grid.dims) // 2)
        self.shift_input = _numexpr_grid_shift(-center, self.output_grid)

        # Remove piston shift (remove central shift phase)
        self.shift_input /= np.fft.ifftshift(self.shift_input.reshape(self.shape_out)).ravel()[0]

        # Calculate the multiplication for emulating the FFTshift (if requested).
        if emulate_fftshifts:
            f_shift = input_grid.delta * np.array(self.internal_shape[::-1] // 2)
            fftshift = _numexpr_grid_shift(f_shift, self.internal_grid)

            if self.cutout_output:
                self.shift_input *= fftshift.reshape(self.internal_shape)[self.cutout_output].ravel()
            else:
                self.shift_input *= fftshift

        # Apply weights for Fourier normalization.
        self.shift_input *= self.weights

        # Calculate the shift array when the output grid was shifted compared to the native shift
        # expcted by the numpy FFT implementation.
        shift = np.ones(self.input_grid.ndim) * shift
        if np.allclose(shift, 0):
            self.shift_output = 1
        else:
            self.shift_output = _numexpr_grid_shift(-shift, self.input_grid)

        # Calculate the multiplication for emulating the FFTshift (if requested).
        if emulate_fftshifts:
            f_shift = self.input_grid.delta * np.array(self.internal_shape[::-1] // 2)
            fftshift = _numexpr_grid_shift(f_shift, self.internal_grid)

            fftshift *= np.exp(-1j * np.dot(f_shift, self.internal_grid.zero))

            if self.cutout_input:
                self.shift_output *= fftshift.reshape(self.internal_shape)[self.cutout_input].ravel()
            else:
                self.shift_output *= fftshift

        # Detect if we don't need to shift in the output plane (to avoid a multiplication in the operation).
        if np.isscalar(self.shift_output) and np.allclose(self.shift_output, 1):
            self.shift_output = None

    @multiplex_for_tensor_fields
    def forward(self, field):
        '''Returns the forward Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to Fourier transform.

        Returns
        -------
        Field
            The Fourier transform of the field.
        '''
        if self.cutout_input is None:
            self.internal_array[:] = field.reshape(self.shape_in)

            if self.shift_output is not None:
                self.internal_array *= self.shift_output.reshape(self.shape_in)
        else:
            self.internal_array[:] = 0
            self.internal_array[self.cutout_input] = field.reshape(self.shape_in)

            if self.shift_output is not None:
                self.internal_array[self.cutout_input] *= self.shift_output.reshape(self.shape_in)

        if not self.emulate_fftshifts:
            self.internal_array = np.fft.ifftshift(self.internal_array)

        fft_array = _fft_module.fftn(self.internal_array)

        if not self.emulate_fftshifts:
            fft_array = np.fft.fftshift(fft_array)

        if self.cutout_output is None:
            res = fft_array.ravel()
        else:
            res = fft_array[self.cutout_output].ravel()

        res *= self.shift_input

        float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
        return Field(res, self.output_grid).astype(complex_dtype, copy=False)

    @multiplex_for_tensor_fields
    def backward(self, field):
        '''Returns the inverse Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to inverse Fourier transform.

        Returns
        -------
        Field
            The inverse Fourier transform of the field.
        '''
        if self.cutout_output is None:
            self.internal_array[:] = field.reshape(self.shape_out)
            self.internal_array /= self.shift_input.reshape(self.shape_out)
        else:
            self.internal_array[:] = 0
            self.internal_array[self.cutout_output] = field.reshape(self.shape_out)
            self.internal_array[self.cutout_output] /= self.shift_input.reshape(self.shape_out)

        if not self.emulate_fftshifts:
            self.internal_array = np.fft.ifftshift(self.internal_array)

        fft_array = _fft_module.ifftn(self.internal_array)

        if not self.emulate_fftshifts:
            fft_array = np.fft.fftshift(fft_array)

        if self.cutout_input is None:
            res = fft_array.ravel()
        else:
            res = fft_array[self.cutout_input].ravel()

        if self.shift_output is not None:
                res /= self.shift_output

        float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
        return Field(res, self.input_grid).astype(complex_dtype, copy=False)



import numpy as np

from ..field import Field
from ..util import SpectralNoiseFactoryFFT, inverse_tikhonov
from .apodization import SurfaceApodizer
from ..propagation import FresnelPropagator
from ..aperture import make_circular_aperture
from .optical_element import OpticalElement

def make_power_law_error(pupil_grid, ptv, diameter, exponent=-2.5, aperture=None, remove_modes=None):
    '''Create an error surface from a power-law power spectral density.

    Parameters
    ----------
    pupil_grid : Grid
        The grid on which to calculate the error.
    ptv : scalar
        The peak-to-valley of the wavefront aberration in meters.
    diameter : scalar
        The diameter over which the ptv is calculated.
    exponent : scalar
        The exponent of the power law.
    aperture : Field
        The mask over which to calculate the ptv. A circular aperture with diameter
        `diameter` is used if this is not given.
    remove_modes : ModeBasis
        The modes which to remove from the surface aberration. The peak-to-valley
        is enforced before these modes are removed. This allows for correctting surface
        errors with optic alignment.

    Returns
    -------
    Field
        The surface error calculated on `pupil_grid`.
    '''
    def psd(grid):
        res = Field(grid.as_('polar').r**exponent, grid)
        res[grid.as_('polar').r == 0] = 0
        return res

    if aperture is None:
        aperture = make_circular_aperture(diameter)(pupil_grid)

    screen = SpectralNoiseFactoryFFT(psd, pupil_grid).make_random()()
    screen *= ptv / np.ptp(screen[aperture != 0])

    if remove_modes is not None:
        trans = remove_modes.transformation_matrix
        trans_inv = inverse_tikhonov(trans, 1e-6)
        screen -= trans.dot(trans_inv.dot(screen))

    return Field(screen * aperture, pupil_grid)

class SurfaceAberration(SurfaceApodizer):
    '''A surface aberration with a specific power law.

    Parameters
    ----------
    pupil_grid : Grid
        The grid on which the incoming wavefront is defined.
    ptv : scalar
        The peak-to-valley of the wavefront aberration in meters.
    diameter : scalar
        The diameter over which the ptv is calculated.
    exponent : scalar
        The exponent of the power law.
    refractive_index : scalar
        The refractive index of the surface for which this is the surface error.
        The default is a mirror surface.
    aperture : Field
        The mask over which to calculate the ptv. A circular aperture with diameter
        `diameter` is used if this is not given.
    remove_modes : ModeBasis
        The modes which to remove from the surface aberration. The peak-to-valley
        is enforced before these modes are removed. This allows for correcting surface
        errors with optic alignment.
    '''
    def __init__(self, pupil_grid, ptv, diameter, exponent=-2.5, refractive_index=-1, aperture=None, remove_modes=None):
        surface = make_power_law_error(pupil_grid, ptv, diameter, exponent, aperture, remove_modes)
        SurfaceApodizer.__init__(self, surface, refractive_index)

class SurfaceAberrationAtDistance(OpticalElement):
    '''A surface at a certain distance from the current plane.

    Light is propagated to this surface, then the surface errors are
    applied, and afterwards the light is propagated back towards the
    original plane. This allows for easy addition of surface errors on
    lenses, while still retaining the Fraunhofer propagations in between
    focal and pupil planes.

    Parameters
    ----------
    surface_aberration : OpticalElement
        The optical element describing the surface aberration.
    distance : scalar
        The distance from the current plane.
    '''
    def __init__(self, surface_aberration, distance):
        self.fresnel = FresnelPropagator(surface_aberration.input_grid, distance)
        self.surface_aberration = surface_aberration

    def forward(self, wavefront):
        '''Propagate a wavefront forwards through the surface aberration.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront after the surface aberration. This wavefront is
            given at the same plane as `wavefront`.
        '''
        wf = self.fresnel.forward(wavefront)
        wf = self.surface_aberration.forward(wf)
        return self.fresnel.backward(wf)

    def backward(self, wavefront):
        '''Propagate a wavefront backwards through the surface aberration.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront before the surface aberration. This wavefront is
            given at the same plane as `wavefront`.
        '''
        wf = self.fresnel.forward(wavefront)
        wf = self.surface_aberration.backward(wf)
        return self.fresnel.backward(wf)



from hcipy import *
import numpy as np

def test_optical_differentiation_wavefront_sensor():
    pupil_grid = make_pupil_grid(128, 1)
    wfs_grid = make_pupil_grid(256, 2)
    amplitude_filter = make_polarization_odwfs_amplitude_filter(0.0)

    odwfs = OpticalDifferentiationWavefrontSensorOptics(amplitude_filter, pupil_grid, wfs_grid)

    zernike_modes = make_zernike_basis(20, 1, pupil_grid)
    aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

    wf = Wavefront(make_circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
    odwfs(wf).intensity

def test_pyramid_wavefront_sensor():
    pupil_grid = make_pupil_grid(128)
    wfs_grid = make_pupil_grid(256, 2)

    pywfs = PyramidWavefrontSensorOptics(pupil_grid, wfs_grid, 1.0)

    zernike_modes = make_zernike_basis(20, 1, pupil_grid)
    aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

    wf = Wavefront(make_circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
    pywfs(wf).intensity

def test_modulated_pyramid_wavefront_sensor():
    pupil_grid = make_pupil_grid(128)
    wfs_grid = make_pupil_grid(256, 2)

    num_steps = 12

    pywfs = PyramidWavefrontSensorOptics(pupil_grid, wfs_grid, 1.0, q=3)
    mpywfs = ModulatedPyramidWavefrontSensorOptics(pywfs, 2, num_steps=num_steps)
    fast_mpywfs = ModulatedPyramidWavefrontSensorOptics(pywfs, 2, num_steps=num_steps, fast_modulation_method=True)

    zernike_modes = make_zernike_basis(20, 1, pupil_grid)
    aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

    wf = Wavefront(make_circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
    modulated_wfs = mpywfs(wf)

    assert len(modulated_wfs) == num_steps

    total_image = 0
    for wi in modulated_wfs:
        total_image += wi.power / num_steps

    total_image_fast = 0
    for wi in fast_mpywfs(wf):
        total_image_fast += wi.power / num_steps

    assert (abs(total_image - total_image_fast).max() / total_image.max()) < 2e-2

def test_zernike_wavefront_sensor():
    pupil_grid = make_pupil_grid(128)

    zwfs = ZernikeWavefrontSensorOptics(pupil_grid)

    zernike_modes = make_zernike_basis(20, 1, pupil_grid)
    aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

    wf = Wavefront(make_circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
    zwfs(wf).intensity

def test_vector_zernike_wavefront_sensor():
    pupil_grid = make_pupil_grid(128)

    vzwfs = VectorZernikeWavefrontSensorOptics(pupil_grid)

    zwfs_plus = ZernikeWavefrontSensorOptics(pupil_grid, np.pi / 2)
    zwfs_neg = ZernikeWavefrontSensorOptics(pupil_grid, -np.pi / 2)

    zernike_modes = make_zernike_basis(20, 1, pupil_grid)
    aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

    wf = Wavefront(make_circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))

    vector_img = vzwfs(wf).intensity
    ref_img = (zwfs_plus(wf).intensity + zwfs_neg(wf).intensity) / 2

    assert np.allclose(vector_img, ref_img)



import pytest
import hcipy
import numpy as np


def _parameters():
    for func in ['fft', 'ifft', 'fft2', 'ifft2', 'fftn', 'ifftn', 'rfft', 'irfft', 'rfft2', 'irfft2', 'rfftn', 'irfftn', 'hfft', 'ihfft']:
        for dtype_in in ['float32', 'float64', 'complex64', 'complex128']:
            real_out = func.startswith('ir') or func.startswith('h')

            if dtype_in == 'float32' or dtype_in == 'complex64':
                dtype_out = 'float32' if real_out else 'complex64'
            else:
                dtype_out = 'float64' if real_out else 'complex128'

            if func.startswith('r') and dtype_in.startswith('complex'):
                continue

            if func.startswith('ih') and dtype_in.startswith('complex'):
                continue

            for method in ['numpy', 'scipy', 'fftw', 'mkl']:
                if ('r' in func or 'h' in func) and method == 'mkl':
                    # MKL doesn't implement rffts or hffts.
                    continue

                yield (func, method, dtype_in, dtype_out)

@pytest.mark.parametrize('func,method,dtype_in,dtype_out', list(_parameters()))
def test_fft_acceleration(func, method, dtype_in, dtype_out):
    rng = np.random.default_rng(seed=0)

    if method == 'fftw':
        pytest.importorskip('pyfftw')

    if method == 'mkl':
        pytest.importorskip('mkl_fft')

    for N in [32, 128, 1024]:
        x = rng.standard_normal((N, N)).astype(dtype_in)

        if dtype_in.startswith('complex'):
            x = x + 1j * rng.standard_normal((N, N)).astype(dtype_in)

        numpy_func = getattr(np.fft, func)
        hcipy_func = getattr(hcipy._math.fft, func)

        y_numpy = numpy_func(x).astype(dtype_out)
        y_method = hcipy_func(x, method=method)

        if dtype_out == 'complex64' or dtype_out == 'float32':
            rtol = 5e-4
        else:
            rtol = 1e-8

        assert np.allclose(y_numpy, y_method, rtol=rtol, atol=rtol)
        assert y_method.dtype == dtype_out



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..propagation import FraunhoferPropagator
from ..optics import MultiplexedComplexSurfaceApodizer
from ..field import make_pupil_grid, Field

import numpy as np

def make_polarization_odwfs_amplitude_filter(beta):
    '''
    The OD-wfs amplitude filter based on polarization optics following [Haffert2016]_.

    .. [Haffert2016] S. Y. Haffert, 2016, "Generalised optical differentiation wavefront sensor:
        a sensitive high dynamic range wavefront sensor," Opt. Express 24, 18986-19007 (2016).

    Parameters
    ----------
    beta : scalar

    Returns
    -------
    function
        A function that generates the amplitude profile.

    Raises
    ------
    ValueError
        If beta is not between 0 and 1.
    '''
    if beta < 0 or beta > 1:
        raise ValueError('Beta should be between 0 and 1.')

    return lambda x: np.sin(np.pi / 2 * (beta * np.heaviside(x, 0.5) + (1 - beta) * (1 + x) / 2))

def make_odwfs_amplitude_filter(beta):
    '''
    The OD-wfs amplitude filter following [Haffert2016]_.

    .. [Haffert2016] S. Y. Haffert, 2016, "Generalised optical differentiation wavefront sensor:
        a sensitive high dynamic range wavefront sensor," Opt. Express 24, 18986-19007 (2016).

    Parameters
    ----------
    beta : scalar

    Returns
    -------
    function
        A function that generates the amplitude profile.

    Raises
    ------
    ValueError
        If beta is not between 0 and 1.
    '''
    if beta < 0 or beta > 1:
        raise ValueError('Beta should be between 0 and 1.')

    return lambda x: (beta * np.heaviside(x, 0.5) + (1 - beta) * (1 + x) / 2) / np.sqrt(2)

def optical_differentiation_surface(filter_size, amplitude_filter, separation, wavelength_0, refractive_index, orientation=0):
    '''A generator function for the complex multiplexed surface of the ODWFS.

    Parameters
    ----------
    filter_size : scalar
        The physical size of the filter in lambda/D.
    amplitude_filter : lambda function
        A lambda function which acts on the focal plane grid to create a amplitude filter.
    separation : scalar
        The separation of the pupils in pupil diameters.
    wavelength_0 : scalar
        The reference wavelength for the filter specifications.
    refractive_index : lambda function
        A lambda function for the refractive index which accepts a wavelength.
    orientation : scalar
        The orientation of the amplitude filters. The default value is 0 radian.
    Returns
    ----------
    function
        The returned function acts on a grid to create the amplitude filters for that grid.
    '''
    def func(grid):
        surface_grid = grid.rotated(orientation)

        surface_tilt = 0.5 * separation / (refractive_index(wavelength_0) - 1)

        # The surfaces which tilt the beam
        surf1 = -surface_tilt * surface_grid.x
        surf2 = -surface_tilt * -surface_grid.x
        surf3 = -surface_tilt * surface_grid.y
        surf4 = -surface_tilt * -surface_grid.y

        surf = (Field(surf1, surface_grid), Field(surf2, surface_grid), Field(surf3, surface_grid), Field(surf4, surface_grid))

        # The physical boundaries of the mask
        filter_mask = (np.abs(grid.x) < filter_size) * (np.abs(grid.y) < filter_size)

        # Note: be careful with the plus and minus signs of the filters.
        # For energy conservation the squared sum of the filters should be <= 1
        # For electric field conservation the second filter has to have opposite sign.
        filter_1 = amplitude_filter(surface_grid.x / filter_size)
        filter_1 *= filter_mask

        filter_2 = -amplitude_filter(-surface_grid.x / filter_size)
        filter_2 *= filter_mask

        filter_3 = amplitude_filter(surface_grid.y / filter_size)
        filter_3 *= filter_mask
        for f in filter_3:
            print(f)

        filter_4 = -amplitude_filter(-surface_grid.y / filter_size)
        filter_4 *= filter_mask

        amp = (Field(filter_1, grid), Field(filter_2, grid), Field(filter_3, grid), Field(filter_4, grid))

        return MultiplexedComplexSurfaceApodizer(amp, surf, refractive_index)
    return func

class OpticalDifferentiationWavefrontSensorOptics(WavefrontSensorOptics):
    '''The optical elements for a optical-differentiation wavefront sensor.

    Parameters
    ----------
    amplitude_filter : callable
        The function that defines the amplitude filter in the focal plane.
    input_grid : Grid
        The grid on which the input wavefront is defined.
    output_grid : Grid
        The grid on which the output wavefront is defined.
    separation : scalar
        The separation between the pupils. The default takes the input grid extent as separation.
    D : scalar
        The size of the pupil. The default take sthe input grid extent as pupil size.
    wavelength_0 : scalar
        The reference wavelength that determines the physical scales.
    q : scalar
        The focal plane oversampling coefficient. The default uses the minimal required sampling.
    num_airy : scalar
        The radius of the focal plane spatial filter in units of lambda/D at the reference wavelength.
    refractive_index : callable
        A callable that returns the refractive index as function of wavelength.
        The default is a refractive index of 1.5.
    '''
    def __init__(self, amplitude_filter, input_grid, output_grid, separation=None, D=None, wavelength_0=1, q=None, num_airy=None, refractive_index=lambda x: 1.5):
        if not input_grid.is_regular:
            raise ValueError('The input grid must be a regular grid.')

        self.input_grid = input_grid
        self.output_grid = output_grid

        if D is None:
            D = np.max(input_grid.delta * (input_grid.shape - 1))

        if separation is None:
            separation = D

        # Create the intermediate focal grid
        # Oversampling necessary to see all frequencies in the output wavefront sensor plane
        qmin = np.ceil(max(output_grid.x.ptp() / input_grid.x.ptp(), 2))
        if q is None:
            q = qmin
        elif q < qmin:
            raise ValueError('The requested focal plane sampling is too low to sufficiently sample the wavefront sensor output.')

        if num_airy is None:
            self.num_airy = np.max(input_grid.shape - 1) / 2
        else:
            self.num_airy = num_airy

        num_pixels = 2 * int(self.num_airy * q)
        spatial_resolution = wavelength_0 / D
        self.focal_grid = make_pupil_grid(num_pixels, 2 * spatial_resolution * self.num_airy)

        # Make all the optical elements
        self.filter_size = self.num_airy * wavelength_0 / D
        focal_plane_mask = optical_differentiation_surface(self.filter_size, amplitude_filter, separation * np.sqrt(2), wavelength_0, refractive_index, orientation=np.pi / 4)
        self.focal_mask = focal_plane_mask(self.focal_grid)

        # Make the propagators
        self.pupil_to_focal = FraunhoferPropagator(self.input_grid, self.focal_grid)
        self.focal_to_pupil = FraunhoferPropagator(self.focal_grid, self.output_grid)

    def forward(self, wavefront):
        '''Propagates a wavefront through the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        Wavefront
            The output wavefront.
        '''
        wf = self.pupil_to_focal.forward(wavefront)
        wf = self.focal_mask.forward(wf)
        wf = self.focal_to_pupil.forward(wf)

        return wf

    def backward(self, wavefront):
        '''Propagates a wavefront backwards through the wavefront sensor.

        Parameters
        ----------
        wavefront : Wavefront
            The input wavefront that will propagate through the system.

        Returns
        -------
        Wavefront
            The output wavefront.
        '''
        wf = self.pupil_to_focal.backward(wavefront)
        wf = self.focal_mask.backward(wf)
        wf = self.focal_to_pupil.backward(wf)

        return wf

class OpticalDifferentiationWavefrontSensorEstimator(WavefrontSensorEstimator):
    '''Estimates the wavefront slopes from pyramid wavefront sensor images.

    Parameters
    ----------
    aperture : function
        A function which mask the pupils for the normalized differences.
    output_grid : Grid
        The grid on which the output of a pyramid wavefront sensor is sampled.

    Attributes
    ----------
    measurement_grid : Grid
        The grid on which the normalized differences are defined.
    pupil_mask : array_like
        A mask for the normalized differences.
    num_measurements : int
        The number of pixels in the output vector.
    '''
    def __init__(self, slope_grid, aperture):
        self.pupil_mask = aperture
        self.num_measurements = 2 * int(np.sum(self.pupil_mask > 0))

    def estimate(self, images):
        '''A function which estimates the wavefront slope from a pyramid image.

        Parameters
        ----------
        images : list
            A list of scalar intensity fields containing pyramid wavefront sensor images.

        Returns
        -------
        array_like
            The wavefront sensor slopes.
        '''
        import warnings
        warnings.warn("This function does not work as expected and will be changed in a future update.", RuntimeWarning)

        image = images.shaped
        sub_shape = np.array(image.grid.shape) // 2

        # Subpupils
        I_a = image[:sub_shape[0], :sub_shape[1]]
        I_b = image[sub_shape[0]:2 * sub_shape[0], :sub_shape[1]]
        I_c = image[sub_shape[0]:2 * sub_shape[0], sub_shape[1]:2 * sub_shape[1]]
        I_d = image[:sub_shape[0], sub_shape[1]:2 * sub_shape[1]]

        I_x = (I_a - I_c) / (I_a + I_c)
        I_y = (I_b - I_d) / (I_b + I_d)

        I_x = I_x.ravel()[self.pupil_mask > 0]
        I_y = I_y.ravel()[self.pupil_mask > 0]

        slopes = np.vstack((I_x, I_y))
        return slopes



import numpy as np
import copy

from ..field import Field
from ..fourier import FastFourierTransform, MatrixFourierTransform

class SpectralNoiseFactory(object):
    def __init__(self, psd, output_grid):
        '''A factory class for spectral noise.

        Parameters
        ----------
        psd : Field generator
            The power spectral density of the noise.
        output_grid : Grid
            The grid on which to compute the noise.
        '''
        self.psd = psd
        self.output_grid = output_grid

    def make_random(self, seed=None):
        '''Make a single realization of the spectral noise.

        This function needs to be implemented in all child classes.

        Parameters
        ----------
        seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
            A seed to initialize the spectral noise. If None, then fresh, unpredictable
            entry will be pulled from the OS. If an int or array of ints, then it will
            be passed to a numpy.SeedSequency to derive the initial BitGenerator state.
            If a BitGenerator or Generator are passed, these will be wrapped and used
            instead. Default: None.

        Returns
        -------
        SpectralNoise
            A realization of the spectral noise, that can be shifted and evaluated.
        '''
        raise NotImplementedError()

class SpectralNoise(object):
    '''A spectral noise object.

    This object should not be used directly, but rather be made by a SpectralNoiseFactory object.
    '''
    def copy(self):
        '''Return a copy.

        Returns
        -------
        SpectralNoise
            A copy of ourselves.
        '''
        return copy.deepcopy(self)

    def shift(self, shift):
        '''In-place shift the noise along the grid axes.

        This function needs to be implemented by the child class.

        Parameters
        ----------
        shift : array_like
            The shift in the grid axes.
        '''
        raise NotImplementedError()

    def shifted(self, shift):
        '''Return a copy, shifted by `shift`.

        Parameters
        ----------
        shift : array_like
            The shift in the grid axes.

        Returns
        -------
        SpectralNoise
            A copy of ourselves, shifted by `shift`.
        '''
        a = self.copy()
        a.shift(shift)

        return a

    def __call__(self):
        '''Evaluate the noise on the pre-specified grid.

        This function should be implemented by all child classes.

        Returns
        -------
        Field
            The computed spectral noise.
        '''
        raise NotImplementedError()

class SpectralNoiseFactoryFFT(SpectralNoiseFactory):
    '''A spectral noise factory based on FFTs.

    The use of FFTs means that these spectral noises will wrap at the
    edge of the grid, i.e. they are continuous from one side of the grid
    to the other end. This means that the lowest frequencies of the PSD
    are not well represented and diminished in amplitude. The sampled
    spatial frequencies can be extended by increasing the oversample
    parameter at the cost of memory usage and computation time.

    See SpectralNoiseFactoryMultiscale for an alternative if this wrapping
    should be avoided at only a minor computational cost.

    Parameters
    ----------
    psd : Field generator
        The power spectral density of the noise.
    output_grid : Grid
        The grid on which to compute the noise.
    oversample : integer
        The amount by which to oversample to grid. For values higher than one,
        the spectral noise can be shifted by that fraction of the grid extent
        without wrapping.
    '''
    def __init__(self, psd, output_grid, oversample=1):
        SpectralNoiseFactory.__init__(self, psd, output_grid)

        if not self.output_grid.is_regular:
            raise ValueError("Can't make a SpectralNoiseFactoryFFT on a non-regular grid.")

        self.fourier = FastFourierTransform(self.output_grid, q=oversample)
        self.input_grid = self.fourier.output_grid

        self.period = output_grid.coords.delta * output_grid.coords.shape

        # * (2 * np.pi)**self.input_grid.ndim is due to conversion from PSD from "per Hertz" to "per radian", which yields a factor of 2pi per dimension
        self.C = np.sqrt(self.psd(self.input_grid) / self.input_grid.weights * (2 * np.pi)**self.input_grid.ndim)

    def make_random(self, seed=None):
        '''Make a single realization of the spectral noise.

        Parameters
        ----------
        seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
            A seed to initialize the spectral noise. If None, then fresh, unpredictable
            entry will be pulled from the OS. If an int or array of ints, then it will
            be passed to a numpy.SeedSequency to derive the initial BitGenerator state.
            If a BitGenerator or Generator are passed, these will be wrapped and used
            instead. Default: None.

        Returns
        -------
        SpectralNoiseFFT
            A realization of the spectral noise, that can be shifted and evaluated.
        '''
        rng = np.random.default_rng(seed)

        N = self.input_grid.size

        C = self.C * (rng.standard_normal(N) + 1j * rng.standard_normal(N))
        C = Field(C, self.input_grid)

        return SpectralNoiseFFT(self, C)

class SpectralNoiseFFT(SpectralNoise):
    '''A single realization of FFT spectral noise.

    Parameters
    ----------
    factory : SpectralNoiseFactoryMultiscale
        The factory used to generate this spectral noise instance.
    C : Field
        The PSD noise realization in Fourier space.
    '''
    def __init__(self, factory, C):
        self.factory = factory
        self.C = C

        self.coords = C.grid.separated_coords

    def shift(self, shift):
        '''In-place shift the noise along the grid axes.

        Parameters
        ----------
        shift : array_like
            The shift in the grid axes.
        '''
        S = [shift[i] * self.coords[i] for i in range(len(self.coords))]
        S = np.add.reduce(np.ix_(*S))

        self.C *= np.exp(-1j * S.ravel())

    def __call__(self):
        '''Evaluate the noise on the pre-specified grid.

        Returns
        -------
        Field
            The computed spectral noise.
        '''
        return self.factory.fourier.backward(self.C).real

class SpectralNoiseFactoryMultiscale(SpectralNoiseFactory):
    '''A spectral noise factory based on multiscale Fourier transforms.

    This factory should provide similar-looking noise compared to
    SpectralNoiseFactoryFFT, at a fraction of the computational cost.
    It does this by performing multiscale FTs rather than full-scale FTs.
    This means that the noise entropy is slightly reduced, however in the
    vast majority of cases completely invisibly so.

    Parameters
    ----------
    psd : Field generator
        The power spectral density of the noise.
    output_grid : Grid
        The grid on which to compute the noise.
    oversampling : integer
        The amount by which to oversample to grid. For values higher than one,
        the spectral noise can be shifted by that fraction of the grid extent
        without wrapping.
    '''
    def __init__(self, psd, output_grid, oversampling):
        SpectralNoiseFactory.__init__(self, psd, output_grid)

        self.oversampling = oversampling

        self.fourier_1 = FastFourierTransform(self.output_grid)
        self.input_grid_1 = self.fourier_1.output_grid

        self.input_grid_2 = self.input_grid_1.scaled(1.0 / oversampling)
        self.fourier_2 = MatrixFourierTransform(self.output_grid, self.input_grid_2)

        boundary = np.abs(self.input_grid_2.x).max()
        mask_1 = self.input_grid_1.as_('polar').r < boundary
        mask_2 = self.input_grid_2.as_('polar').r >= boundary

        # * (2*np.pi)**self.input_grid.ndim is due to conversion from PSD from "per Hertz" to "per radian", which yields a factor of 2pi per dimension
        self.C_1 = np.sqrt(psd(self.input_grid_1) / self.input_grid_1.weights * (2 * np.pi)**self.input_grid_1.ndim)
        self.C_1[mask_1] = 0
        self.C_2 = np.sqrt(psd(self.input_grid_2) / self.input_grid_2.weights * (2 * np.pi)**self.input_grid_1.ndim)
        self.C_2[mask_2] = 0

    def make_random(self, seed=None):
        '''Make a single realization of the spectral noise.

        Parameters
        ----------
        seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
            A seed to initialize the spectral noise. If None, then fresh, unpredictable
            entry will be pulled from the OS. If an int or array of ints, then it will
            be passed to a numpy.SeedSequency to derive the initial BitGenerator state.
            If a BitGenerator or Generator are passed, these will be wrapped and used
            instead. Default: None.

        Returns
        -------
        SpectralNoiseMultiscale
            A realization of the spectral noise, that can be shifted and evaluated.
        '''
        rng = np.random.default_rng(seed)

        N_1 = self.input_grid_1.size
        N_2 = self.input_grid_2.size

        C_1 = self.C_1 * (rng.standard_normal(N_1) + 1j * rng.standard_normal(N_1))
        C_2 = self.C_2 * (rng.standard_normal(N_2) + 1j * rng.standard_normal(N_2))

        return SpectralNoiseMultiscale(self, C_1, C_2)

class SpectralNoiseMultiscale(SpectralNoise):
    '''A single realization of multiscale spectral noise.

    Parameters
    ----------
    factory : SpectralNoiseFactoryMultiscale
        The factory used to generate this spectral noise instance.
    C_1 : Field
        The high-frequency part of the PSD noise realization.
    C_2 : Field
        The low-frequency part of the PSD noise realization.
    '''
    def __init__(self, factory, C_1, C_2):
        self.factory = factory

        self.C_1 = C_1
        self.C_2 = C_2

        self.coords_1 = C_1.grid.separated_coords
        self.coords_2 = C_2.grid.separated_coords

    def shift(self, shift):
        '''In-place shift the noise along the grid axes.

        Parameters
        ----------
        shift : array_like
            The shift in the grid axes.
        '''
        S_1 = [shift[i] * self.coords_1[i] for i in range(len(self.coords_1))]
        S_1 = sum(np.ix_(*S_1))

        S_2 = [shift[i] * self.coords_2[i] for i in range(len(self.coords_2))]
        S_2 = sum(np.ix_(*S_2))

        self.C_1 *= np.exp(-1j * S_1.ravel())
        self.C_2 *= np.exp(-1j * S_2.ravel())

    def __call__(self):
        '''Evaluate the noise on the pre-specified grid.

        Returns
        -------
        Field
            The computed spectral noise.
        '''
        ps = self.factory.fourier_1.backward(self.C_1).real
        ps += self.factory.fourier_2.backward(self.C_2).real

        return ps



import matplotlib
matplotlib.use('Agg')

from hcipy import *
import numpy as np
import matplotlib.pyplot as plt
import os
import pytest
import shutil

def is_ffmpeg_installed():
    ffmpeg_path = Configuration().plotting.ffmpeg_path

    if ffmpeg_path is None:
        ffmpeg_path = 'ffmpeg'

    return shutil.which(ffmpeg_path) is not None

def check_animation(mw, style):
    grid = make_pupil_grid(256)

    for i in range(25):
        field = Field(np.random.randn(grid.size), grid)

        if style == 'mpl':
            plt.clf()
            imshow_field(field)

            mw.add_frame()
        elif style == 'fig':
            fig, ax = plt.subplots()
            imshow_field(field, ax=ax)

            mw.add_frame(fig=fig)
            plt.close(fig)
        elif style == 'img':
            mw.add_frame(data=field.shaped, cmap='RdBu')
        else:
            raise ValueError('Animation style not known.')

    mw.close()

    pytest.raises(RuntimeError, mw.add_frame)

@pytest.mark.parametrize('style', ['mpl', 'fig', 'img'])
def test_frame_writer(style):
    mw = FrameWriter('test_frames/')

    check_animation(mw, style=style)

    assert os.path.isdir('test_frames')
    shutil.rmtree('test_frames')

@pytest.mark.parametrize('style', ['mpl', 'fig', 'img'])
def test_gif_writer(style):
    mw = GifWriter('test.gif')

    check_animation(mw, style=style)

    assert os.path.isfile('test.gif')
    os.remove('test.gif')

@pytest.mark.parametrize('style', ['mpl', 'fig', 'img'])
@pytest.mark.skipif(not is_ffmpeg_installed(), reason='FFMpeg is not installed.')
def test_ffmpeg_writer(style):
    mw = FFMpegWriter('test.mp4')

    check_animation(mw, style=style)

    assert mw._repr_html_()

    assert os.path.isfile('test.mp4')
    os.remove('test.mp4')

def test_imshow_field():
    grid = make_pupil_grid(256)

    field = Field(np.random.randn(grid.size), grid)

    imshow_field(field)
    plt.clf()

    mask = make_circular_aperture(1)(grid)

    imshow_field(field, mask=mask)
    plt.clf()

    field = Field(np.random.randn(grid.size) + 1j * np.random.randn(grid.size), grid)

    imshow_field(field)
    plt.clf()

def test_imsave_field():
    grid = make_pupil_grid(256)

    field = Field(np.random.randn(grid.size), grid)

    imsave_field('field.png', field)
    assert os.path.isfile('field.png')

    os.remove('field.png')

def test_contour_field():
    grid = make_pupil_grid(256)

    field = Field(np.random.randn(grid.size), grid)

    contour_field(field)
    plt.clf()

    contourf_field(field)
    plt.clf()

def test_imshow_util():
    pupil_grid = make_pupil_grid(128)
    focal_grid = make_focal_grid(4, 16)

    aperture = make_magellan_aperture(True)(pupil_grid)
    prop = FraunhoferPropagator(pupil_grid, focal_grid)

    wf = Wavefront(aperture)
    wf.electric_field *= np.exp(0.1j * zernike(6, 2, radial_cutoff=False)(pupil_grid))

    imshow_pupil_phase(wf, remove_piston=True, crosshairs=True, title='phase')
    plt.clf()

    img = prop(wf)

    imshow_psf(img, colorbar_orientation='vertical', normalization='peak', crosshairs=True, title='psf')
    plt.clf()

    imshow_psf(img, scale='linear', colorbar_orientation='vertical', normalization='peak', crosshairs=True, title='psf')
    plt.clf()



import numpy as np
from scipy.signal import windows

from .multi_scale import MultiScaleCoronagraph
from ..optics import LinearRetarder, Apodizer, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward, Wavefront
from ..propagation import FraunhoferPropagator
from ..field import make_focal_grid, Field, field_dot
from ..aperture import make_circular_aperture
from ..fourier import FastFourierTransform, MatrixFourierTransform, FourierFilter

class VortexCoronagraph(MultiScaleCoronagraph):
    def __init__(self, input_grid, charge, lyot_stop=None, q=1024, scaling_factor=4, window_size=32):
        complex_mask = lambda grid: Field(np.exp(1j * charge * grid.as_('polar').theta), grid) * (1 - make_circular_aperture(1e-9)(grid))
        super().__init__(input_grid, complex_mask, lyot_stop, q, scaling_factor, window_size)

class VectorVortexCoronagraph(AgnosticOpticalElement):
    '''A vector vortex coronagraph.

    This :class:`OpticalElement` simulations the propagation of light through
    a vector vortex in the focal plane. To resolve the singularity of this vortex
    phase plate, a multi-scale approach is made. Discretisation errors made at
    a certain level are corrected by the next level with finer sampling.

    Parameters
    ----------
    charge : integer
        The charge of the vortex.
    lyot_stop : Field or OpticalElement
        The Lyot stop for the coronagraph. If it's a Field, it is converted to an
        OpticalElement for convenience. If this is None (default), then no Lyot stop is used.
    phase_retardation : scalar or function
        The phase retardation of the vector vortex plate, potentially as a
        function of wavelength. Changes of the phase retardation as a function
        of spatial position is not yet supported.
    q : scalar
        The minimum number of pixels per lambda/D. The number of levels in the multi-scale
        Fourier transforms will be chosen to reach at least this number of samples. The required
        q for a high-accuracy vortex coronagraph depends on the charge of the vortex. For charge 2,
        this can be as low as 32, but for charge 8 you need ~1024. Lower values give higher performance
        as a smaller number of levels is needed, but increases the sampling errors near the singularity.
        Charges not divisible by four require a much lower q. The default (q=1024) is conservative in
        most cases.
    scaling_factor : scalar
        The fractional increase in spatial frequency sampling per level. Larger scaling factors
        require a smaller number of levels, but each level requires a slower Fourier transform.
        Factors of 2 or 4 usually perform the best.
    window_size : integer
        The size of the next level in the number of pixels on the current layer. Lowering this
        increases performance in exchange for accuracy. Values smaller than 4-8 are not recommended.
    '''
    def __init__(self, charge, lyot_stop=None, phase_retardation=np.pi, q=1024, scaling_factor=4, window_size=32):
        self.charge = charge

        if hasattr(lyot_stop, 'forward') or lyot_stop is None:
            self.lyot_stop = lyot_stop
        else:
            self.lyot_stop = Apodizer(lyot_stop)

        self.phase_retardation = phase_retardation

        self.q = q
        self.scaling_factor = scaling_factor
        self.window_size = window_size

        AgnosticOpticalElement.__init__(self)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        pupil_diameter = input_grid.shape * input_grid.delta

        levels = int(np.ceil(np.log(self.q / 2) / np.log(self.scaling_factor))) + 1
        qs = [2 * self.scaling_factor**i for i in range(levels)]
        num_airys = [input_grid.shape / 2]

        focal_grids = []
        instance_data.props = []
        instance_data.jones_matrices = []

        for i in range(1, levels):
            num_airys.append(num_airys[i - 1] * self.window_size / (2 * qs[i - 1] * num_airys[i - 1]))

        for i in range(levels):
            q = qs[i]
            num_airy = num_airys[i]

            focal_grid = make_focal_grid(q, num_airy, pupil_diameter=pupil_diameter, reference_wavelength=1, focal_length=1)

            fast_axis_orientation = Field(self.charge / 2 * focal_grid.as_('polar').theta, focal_grid)
            retardance = self.evaluate_parameter(self.phase_retardation, input_grid, output_grid, wavelength)

            focal_mask_raw = LinearRetarder(retardance, fast_axis_orientation)
            jones_matrix = focal_mask_raw.jones_matrix

            jones_matrix *= 1 - make_circular_aperture(1e-9)(focal_grid)

            if i != levels - 1:
                wx = windows.tukey(self.window_size, 1, False)
                wy = windows.tukey(self.window_size, 1, False)
                w = np.outer(wy, wx)

                w = np.pad(w, (focal_grid.shape - w.shape) // 2, 'constant').ravel()
                jones_matrix *= 1 - w

            for j in range(i):
                fft = FastFourierTransform(focal_grids[j])
                mft = MatrixFourierTransform(focal_grid, fft.output_grid)

                jones_matrix -= mft.backward(fft.forward(instance_data.jones_matrices[j]))

            if i == 0:
                prop = FourierFilter(input_grid, jones_matrix, q)
            else:
                prop = FraunhoferPropagator(input_grid, focal_grid)

            focal_grids.append(focal_grid)
            instance_data.jones_matrices.append(jones_matrix)
            instance_data.props.append(prop)

    def get_input_grid(self, output_grid, wavelength):
        '''Get the input grid for a specified output grid and wavelength.

        This optical element only supports propagation to the same plane as
        its input.

        Parameters
        ----------
        output_grid : Grid
            The output grid of the optical element.
        wavelength : scalar or None
            The wavelength of the outgoing light.

        Returns
        -------
        Grid
            The input grid corresponding to the output grid and wavelength combination.
        '''
        return output_grid

    def get_output_grid(self, input_grid, wavelength):
        '''Get the output grid for a specified input grid and wavelength.

        This optical element only supports propagation to the same plane as
        its input.

        Parameters
        ----------
        input_grid : Grid
            The input grid of the optical element.
        wavelength : scalar or None
            The wavelength of the incoming light.

        Returns
        -------
        Grid
            The output grid corresponding to the input grid and wavelength combination.
        '''
        return input_grid

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        '''Propagate a wavefront through the vortex coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate. This wavefront is expected to be
            in the pupil plane.

        Returns
        -------
        Wavefront
            The Lyot plane wavefront.
        '''
        wavelength = wavefront.wavelength
        wavefront.wavelength = 1

        for i, (jones_matrix, prop) in enumerate(zip(instance_data.jones_matrices, instance_data.props)):
            if i == 0:
                if not wavefront.is_polarized:
                    wf = Wavefront(wavefront.electric_field, input_stokes_vector=(1, 0, 0, 0))
                else:
                    wf = wavefront

                lyot = Wavefront(prop.forward(wf.electric_field), input_stokes_vector=wf.input_stokes_vector)
            else:
                focal = prop(wavefront)

                if not focal.is_polarized:
                    focal = Wavefront(focal.electric_field, input_stokes_vector=(1, 0, 0, 0))

                focal.electric_field = field_dot(jones_matrix, focal.electric_field)
                if i == 0:
                    lyot = prop.backward(focal)
                else:
                    lyot.electric_field += prop.backward(focal).electric_field

        lyot.wavelength = wavelength
        wavefront.wavelength = wavelength

        if self.lyot_stop is not None:
            lyot = self.lyot_stop.forward(lyot)

        return lyot

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        '''Propagate backwards through the vortex coronagraph.

        This essentially is a forward propagation through the same vortex
        coronagraph, but with the sign of its charge flipped.

        Parameters
        ----------
        wavefront : Wavefront
            The Lyot plane wavefront.

        Returns
        -------
        Wavefront
            The pupil-plane wavefront.
        '''
        if self.lyot_stop is not None:
            wavefront = self.lyot_stop.backward(wavefront)

        wavelength = wavefront.wavelength
        wavefront.wavelength = 1

        for i, (jones_matrix, prop) in enumerate(zip(instance_data.jones_matrices, instance_data.props)):
            if i == 0:
                if not wavefront.is_polarized:
                    wavefront = Wavefront(wavefront.electric_field, input_stokes_vector=(1, 0, 0, 0))
                pup = Wavefront(
                    prop.backward(wavefront.electric_field),
                    input_stokes_vector=wavefront.input_stokes_vector
                )
            else:
                focal = prop(wavefront)
                if not focal.is_polarized:
                    focal = Wavefront(focal.electric_field, input_stokes_vector=(1, 0, 0, 0))

                focal.electric_field = field_dot(jones_matrix.conj(), focal.electric_field)
                pup.electric_field += prop.backward(focal).electric_field

        pup.wavelength = wavelength
        wavefront.wavelength = wavelength

        return pup

def make_ravc_masks(central_obscuration, charge=2, pupil_diameter=1, lyot_undersize=0):
    '''Make field generators for the pupil and Lyot-stop masks for a
    ring apodized vortex coronagraph.

    The formulas were implemented according to [Mawet2013]_.

    .. [Mawet2013] Dimitri Mawet et al. 2013 "Ring-apodized vortex coronagraphs for obscured telescopes. I. Transmissive
        ring apodizers" The Astrophysical Journal Supplement Series 209.1 (2013): 7

    Parameters
    ----------
    central_obscuration : scalar
        The diameter of the central obscuration.
    charge : integer
        The charge of the vortex coronagraph used.
    pupil_diameter : scalar
        The diameter of the pupil.
    lyot_undersize : scalar
        The fraction of the pupil diameter to which to undersize the Lyot stop.

    Returns
    -------
    pupil_mask : Field generator
        The complex transmission of the pupil mask.
    lyot_mask : Field generator
        The complex transmission of the Lyot-stop mask.
    '''
    R0 = central_obscuration / pupil_diameter

    if charge == 2:
        t1 = 1 - 0.25 * (R0**2 + R0 * np.sqrt(R0**2 + 8))
        R1 = R0 / np.sqrt(1 - t1)

        pupil1 = make_circular_aperture(pupil_diameter)
        pupil2 = make_circular_aperture(pupil_diameter * R1)
        co = make_circular_aperture(central_obscuration)
        pupil_mask = lambda grid: (pupil1(grid) * t1 + pupil2(grid) * (1 - t1)) * (1 - co(grid))

        lyot1 = make_circular_aperture(pupil_diameter * R1 + pupil_diameter * lyot_undersize)
        lyot2 = make_circular_aperture(pupil_diameter * (1 - lyot_undersize))
        lyot_stop = lambda grid: lyot2(grid) - lyot1(grid)
    elif charge == 4:
        R1 = np.sqrt(np.sqrt(R0**2 * (R0**2 + 4)) - 2 * R0**2)
        R2 = np.sqrt(R1**2 + R0**2)
        t1 = 0
        t2 = (R1**2 - R0**2) / (R1**2 + R0**2)

        pupil1 = make_circular_aperture(pupil_diameter)
        pupil2 = make_circular_aperture(pupil_diameter * R1)
        pupil3 = make_circular_aperture(pupil_diameter * R2)
        co = make_circular_aperture(central_obscuration)

        pupil_mask = lambda grid: (pupil1(grid) * t2 + pupil3(grid) * (t1 - t2) + pupil2(grid) * (1 - t1)) * (1 - co(grid))

        lyot1 = make_circular_aperture(pupil_diameter * R2 + pupil_diameter * lyot_undersize)
        lyot2 = make_circular_aperture(pupil_diameter * (1 - lyot_undersize))
        lyot_stop = lambda grid: lyot2(grid) - lyot1(grid)
    else:
        raise NotImplementedError()

    return pupil_mask, lyot_stop

def get_ravc_planet_transmission(central_obscuration_ratio, charge=2):
    '''Get the planet transmission for a ring-apodized vortex coronagraph.

    The formulas were implemented according to [Mawet2013]_.

    .. [Mawet2013] Dimitri Mawet et al. 2013 "Ring-apodized vortex coronagraphs for obscured telescopes. I. Transmissive
        ring apodizers" The Astrophysical Journal Supplement Series 209.1 (2013): 7

    Parameters
    ----------
    central_obscuration_ratio : scalar
        The ratio of the central obscuration diameter and the pupil diameter.
    charge : integer
        The charge of the vortex coronagraph used.

    Returns
    -------
    scalar
        The intensity transmission for a sufficiently off-axis point source
        for the ring-apodized vortex coronagraph. Point sources close to the vortex
        singularity will be lower in intensity.
    '''
    R0 = central_obscuration_ratio

    if charge == 2:
        t1_opt = 1 - 0.25 * (R0**2 + R0 * np.sqrt(R0**2 + 8))
        R1_opt = R0 / np.sqrt(1 - t1_opt)

        return t1_opt**2 * (1 - R1_opt**2) / (1 - (R0**2))
    elif charge == 4:
        R1 = np.sqrt(np.sqrt(R0**2 * (R0**2 + 4)) - 2 * R0**2)
        R2 = np.sqrt(R1**2 + R0**2)
        t2 = (R1**2 - R0**2) / (R1**2 + R0**2)

        return t2**2 * (1 - R2**2) / (1 - R0**2)
    else:
        raise NotImplementedError()



import numpy as np

from ..optics import Apodizer, OpticalElement
from ..propagation import FraunhoferPropagator

class LyotCoronagraph(OpticalElement):
    '''A Lyot coronagraph with a small focal-plane mask.

    The area outside of this focal-plane mask is assumed to be fully transmisive. The
    method for propagation is based on [Soummer2007]_.

    .. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
        coronagraph propagation".

    Notes
    -----
    The :class:`LyotCoronagraph` tries to automatically get the grid from the focal_plane_mask.
    However, a bug in AgnosticOpticalElements creates an invalid access which crashes the initialization.
    The correct grid can be passed explicitely to circumvent this bug by using the focal_plane_mask_grid parameter.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the incoming wavefront is defined.
    focal_plane_mask : Field or OpticalElement
        The (complex) transmission of the focal-plane mask. If this is an :class:`OpticalElement`,
        this will be used instead. This allows for more realistic implementations of focal-plane
        masks.
    lyot_stop : Field or OpticalElement or None
        The (complex) transmission of the Lyot stop. If this is an :class:`OpticalElement`,
        this will be used instead. This allows for more realistic implementations of Lyot stops.
    focal_length : scalar
        The internal focal length of the Lyot system.
    focal_plane_mask_grid : Grid
        The grid on which the focal plane mask is defined. If this is none,
        the grid will be determined from the focal plane mask. The default value is None.
    '''
    def __init__(self, input_grid, focal_plane_mask, lyot_stop=None, focal_length=1, focal_plane_mask_grid=None):
        if hasattr(focal_plane_mask, 'input_grid'):
            # Focal plane mask is an optical element.
            if focal_plane_mask_grid is None:
                grid = focal_plane_mask.apodization.grid
            else:
                grid = focal_plane_mask_grid

            self.focal_plane_mask = focal_plane_mask
        else:
            # Focal plane mask is a field.
            if focal_plane_mask_grid is None:
                grid = focal_plane_mask.grid
            else:
                grid = focal_plane_mask_grid

            self.focal_plane_mask = Apodizer(focal_plane_mask)

        if lyot_stop is not None and not hasattr(lyot_stop, 'input_grid'):
            lyot_stop = Apodizer(lyot_stop)
        self.lyot_stop = lyot_stop

        self.prop = FraunhoferPropagator(input_grid, grid, focal_length=focal_length)

    def forward(self, wavefront):
        '''Propagate the wavefront through the Lyot coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate. This wavefront is assumed to be in the pupil plane.

        Returns
        -------
        Wavefront
            The Lyot-plane wavefront.
        '''
        wf_foc = self.prop.forward(wavefront)
        wf_foc.electric_field -= self.focal_plane_mask.forward(wf_foc).electric_field

        lyot = self.prop.backward(wf_foc)

        # The next line is equal to
        #     lyot.electric_field = wavefront.electric_field - lyot.electric_field
        # but doesn't copy a numpy array.
        np.subtract(wavefront.electric_field, lyot.electric_field, out=lyot.electric_field)

        if self.lyot_stop is not None:
            lyot = self.lyot_stop.forward(lyot)

        return lyot

    def backward(self, wavefront):
        '''Propagate the wavefront from the Lyot plane to the pupil plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The pupil-plane wavefront.
        '''
        if self.lyot_stop is not None:
            wf = self.lyot_stop.backward(wavefront)
        else:
            wf = wavefront

        wf_foc = self.prop.forward(wf)
        wf_foc.electric_field -= self.focal_plane_mask.backward(wf_foc).electric_field

        pup = self.prop.backward(wf_foc)

        # The next line is equal to
        #     pup.electric_field = wf.electric_field - pup.electric_field
        # but doesn't copy a numpy array.
        np.subtract(wf.electric_field, pup.electric_field, out=pup.electric_field)

        return pup

class OccultedLyotCoronagraph(OpticalElement):
    '''A Lyot coronagraph with a focal-plane mask.

    The area outside of this focal-plane mask is assumed to be fully absorbing.

    Notes
    -----
    The :class:`OccultedLyotCoronagraph` tries to automatically get the grid from the focal_plane_mask.
    However, a bug in AgnosticOpticalElements creates an invalid access which crashes the initialization.
    The correct grid can be passed explicitely to circumvent this bug by using the focal_plane_mask_grid parameter.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the incoming wavefront is defined.
    focal_plane_mask : Field or OpticalElement
        The (complex) transmission of the focal-plane mask. If this is an :class:`OpticalElement`,
        this will be used instead. This allows for more realistic implementations of focal-plane
        masks.
    lyot_stop : Field or OpticalElement or None
        The (complex) transmission of the Lyot stop. If this is an :class:`OpticalElement`,
        this will be used instead. This allows for more realistic implementations of Lyot stops.
    focal_length : scalar
        The internal focal length of the Lyot system.
    focal_plane_mask_grid : Grid
        The grid on which the focal plane mask is defined. If this is none,
        the grid will be determined from the focal plane mask. The default value is None.
    '''
    def __init__(self, input_grid, focal_plane_mask, focal_length=1, focal_plane_mask_grid=None):

        if hasattr(focal_plane_mask, 'input_grid'):
            # Focal plane mask is an optical element.
            if focal_plane_mask_grid is None:
                grid = focal_plane_mask.apodization.grid
            else:
                grid = focal_plane_mask_grid

            self.focal_plane_mask = focal_plane_mask
        else:
            # Focal plane mask is a field.
            if focal_plane_mask_grid is None:
                grid = focal_plane_mask.grid
            else:
                grid = focal_plane_mask_grid

            self.focal_plane_mask = Apodizer(focal_plane_mask)

        self.prop = FraunhoferPropagator(input_grid, grid, focal_length=focal_length)

    def forward(self, wavefront):
        '''Propagate the wavefront through the Lyot coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate. This wavefront is assumed to be in the pupil plane.

        Returns
        -------
        Wavefront
            The Lyot-plane wavefront.
        '''
        return self.prop.backward(self.focal_plane_mask.forward(self.prop.forward(wavefront)))

    def backward(self, wavefront):
        '''Propagate the wavefront from the Lyot plane to the pupil plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        Wavefront
            The pupil-plane wavefront.
        '''
        return self.prop.backward(self.focal_plane_mask.backward(self.prop.forward(wavefront)))



import numpy as np
from .apodization import SurfaceApodizer
from .optical_element import OpticalElement
from .surface_profiles import spherical_surface_sag, even_aspheric_surface_sag
from ..field import Field

def closest_points(point_grid, evaluated_grid):
    from scipy import spatial

    tree = spatial.KDTree(point_grid.points)
    d, i = tree.query(evaluated_grid.points)

    return Field(i, evaluated_grid), Field(d, evaluated_grid)

class MicroLensArray(OpticalElement):
    '''A parabolic micro-lens array.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the micro-lens array is evaluated.
    lenslet_grid : Grid
        A grid containing the lenslet positions.
    focal_length : scalar
        The focal length of the micro-lenses
    lenslet_shape : field generator
        The shape of a lenslet.
    '''
    def __init__(self, input_grid, lenslet_grid, focal_length, lenslet_shape=None):
        self.input_grid = input_grid
        self.focal_length = focal_length

        self.mla_grid = lenslet_grid

        if lenslet_shape is None:
            indices, distances = closest_points(lenslet_grid, input_grid)

            self.mla_index = indices
            self.mla_opd = (-1 / (2 * focal_length)) * distances**2
        else:
            self.mla_index = Field(-np.ones(self.input_grid.size), self.input_grid)
            self.mla_opd = Field(np.zeros(self.input_grid.size), self.input_grid)

            for i, (x, y) in enumerate(lenslet_grid.as_('cartesian').points):
                shifted_grid = input_grid.shifted((x, y))
                mask = lenslet_shape(shifted_grid) != 0

                self.mla_opd[mask] = (-1 / (2 * focal_length)) * (shifted_grid.x[mask]**2 + shifted_grid.y[mask]**2)
                self.mla_index[mask] = i

        self.mla_surface = SurfaceApodizer(self.mla_opd, 2)

    def forward(self, wavefront):
        return self.mla_surface.forward(wavefront)

    def backward(self, wavefront):
        return self.mla_surface.backward(wavefront)


class SphericalMicroLensArray(OpticalElement):
    '''An even asphere micro-lens array.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the micro-lens array is evaluated.
    lenslet_grid : Grid
        A grid containing the lenslet positions.
    radius_of_curvature : scalar
        The radius of curvature of the micro-lenses.
    lenslet_shape : field generator
        The shape of a lenslet.
    refractive_index : scalar or function
        The refractive index of the material.
    conic_constant : scalar
        The conic constant of the micro-lenses.
    aspheric_coefficients : array_like
        The aspheric coefficients of the micro-lenses.
    '''
    def __init__(self, input_grid, lenslet_grid, radius_of_curvature, lenslet_shape, refractive_index=1.5):

        self.input_grid = input_grid
        self.mla_grid = lenslet_grid
        self.n = refractive_index
        self.radius_of_curvature = radius_of_curvature

        self.mla_index = -self.input_grid.ones()
        self.mla_opd = self.input_grid.zeros()
        self.surface_sag = spherical_surface_sag(radius_of_curvature)

        for i, (x, y) in enumerate(lenslet_grid.as_('cartesian').points):
            shifted_grid = input_grid.shifted((x, y))
            mask = lenslet_shape(shifted_grid) != 0
            subset_grid = shifted_grid.subset(mask)

            # Check if the micro-lens is within the field of view.
            if np.count_nonzero(mask) > 0:
                self.mla_index[mask] = i
                self.mla_opd[mask] += self.surface_sag(subset_grid)

        self.mla_surface = SurfaceApodizer(self.mla_opd, refractive_index)

    def forward(self, wavefront):
        return self.mla_surface.forward(wavefront)

    def backward(self, wavefront):
        return self.mla_surface.backward(wavefront)

class EvenAsphereMicroLensArray(OpticalElement):
    '''An even asphere micro-lens array.

    Parameters
    ----------
    input_grid : Grid
        The grid on which the micro-lens array is evaluated.
    lenslet_grid : Grid
        A grid containing the lenslet positions.
    radius_of_curvature : scalar
        The radius of curvature of the micro-lenses.
    lenslet_shape : field generator
        The shape of a lenslet.
    refractive_index : scalar or function
        The refractive index of the material.
    conic_constant : scalar
        The conic constant of the micro-lenses.
    aspheric_coefficients : array_like
        The aspheric coefficients of the micro-lenses.
    '''
    def __init__(self, input_grid, lenslet_grid, radius_of_curvature, lenslet_shape, refractive_index=1.5, conic_constant=0, aspheric_coefficients=None):
        self.input_grid = input_grid
        self.mla_grid = lenslet_grid
        self.n = refractive_index
        self.radius_of_curvature = radius_of_curvature
        self.conic_constant = conic_constant
        if aspheric_coefficients is None:
            self.aspheric_coefficients = []
        else:
            self.aspheric_coefficients = aspheric_coefficients

        self.mla_index = -self.input_grid.ones()
        self.mla_opd = self.input_grid.zeros()
        self.surface_sag = even_aspheric_surface_sag(radius_of_curvature, conic_constant, aspheric_coefficients)

        for i, (x, y) in enumerate(lenslet_grid.as_('cartesian').points):
            shifted_grid = input_grid.shifted((x, y))
            mask = lenslet_shape(shifted_grid) != 0
            subset_grid = shifted_grid.subset(mask)

            # Check if the micro-lens is within the field of view.
            if np.count_nonzero(mask) > 0:
                self.mla_index[mask] = i
                self.mla_opd[mask] += self.surface_sag(subset_grid)

        self.mla_surface = SurfaceApodizer(self.mla_opd, refractive_index)

    def forward(self, wavefront):
        return self.mla_surface.forward(wavefront)

    def backward(self, wavefront):
        return self.mla_surface.backward(wavefront)



import numpy as np

from ..optics import Wavefront, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward
from ..field import Field, evaluate_supersampled
from ..fourier import FastFourierTransform, make_fft_grid, FourierFilter

class AngularSpectrumPropagator(AgnosticOpticalElement):
    '''The monochromatic angular spectrum propagator for scalar fields.

    The scalar Angular Spectrum propagator is implemented as described by
    [McLeod2014]_. The propagation of an electric field can be described as a transfer
    function in frequency space. The transfer function is taken from
    equation 9 of [McLeod2014]_, and the related impulse response is taken from
    equation 6 of [McLeod2014]_.

    .. [McLeod2014] Robert R. McLeod and Kelvin H. Wagner 2014, "Vector Fourier optics of
        anisotropic materials," Adv. Opt. Photon. 6, 368-412 (2014)

    Parameters
    ----------
    input_grid : anything
        This argument is ignored. The input grid is taken from the incoming wavefront.
    distance : scalar
        The distance to propagate
    num_oversampling : int
        The number of times the transfer function is oversampled. Default is 2.
    wavelength : scalar
        The wavelength of the wavefront.
    refractive_index : scalar
        The refractive index of the medium that the wavefront is propagating in.

    Raises
    ------
    ValueError
        If the `input_grid` is not regular and Cartesian.
    '''
    def __init__(self, input_grid, distance, num_oversampling=2, refractive_index=1):
        self._distance = distance

        self._num_oversampling = num_oversampling
        self._refractive_index = refractive_index

        AgnosticOpticalElement.__init__(self, grid_dependent=True, wavelength_dependent=True)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        if not input_grid.is_regular or not input_grid.is_('cartesian'):
            raise ValueError('The input grid must be a regular, Cartesian grid.')

        k = 2 * np.pi / wavelength * self.evaluate_parameter(self.refractive_index, input_grid, output_grid, wavelength)
        L_max = np.max(input_grid.dims * input_grid.delta)

        if np.any(input_grid.delta < wavelength * abs(self.distance) / L_max):
            def transfer_function(fourier_grid):
                enlarged_grid = make_fft_grid(fourier_grid)
                fft_upscale = FastFourierTransform(enlarged_grid)

                def impulse_response(grid):
                    r_squared = grid.x**2 + grid.y**2 + self.distance**2
                    r = np.sqrt(r_squared)
                    cos_theta = self.distance / r

                    return Field(cos_theta / (2 * np.pi) * np.exp(1j * k * r) * (1 / r_squared - 1j * k / r), grid)

                impulse_response = evaluate_supersampled(impulse_response, enlarged_grid, self.num_oversampling)

                return fft_upscale.forward(impulse_response)
        else:
            def transfer_function_native(fourier_grid):
                k_squared = fourier_grid.as_('polar').r**2
                k_z = np.sqrt(k**2 - k_squared + 0j)

                return Field(np.exp(1j * k_z * self.distance), fourier_grid)

            def transfer_function(fourier_grid):
                return evaluate_supersampled(transfer_function_native, fourier_grid, self.num_oversampling)

        instance_data.fourier_filter = FourierFilter(input_grid, transfer_function, q=2)

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, distance):
        self._distance = distance

        self.clear_cache()

    @property
    def num_oversampling(self):
        return self._num_oversampling

    @num_oversampling.setter
    def num_oversampling(self, num_oversampling):
        self._num_oversampling = num_oversampling

        self.clear_cache()

    @property
    def refractive_index(self):
        return self._refractive_index

    @refractive_index.setter
    def refractive_index(self, refractive_index):
        self._refractive_index = refractive_index

        self.clear_cache()

    def get_input_grid(self, output_grid, wavelength):
        return output_grid

    def get_output_grid(self, input_grid, wavelength):
        return input_grid

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        '''Propagate a wavefront forward by a certain distance.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront after the propagation.
        '''
        filtered = instance_data.fourier_filter.forward(wavefront.electric_field)

        return Wavefront(filtered, wavefront.wavelength, wavefront.input_stokes_vector)

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        '''Propagate a wavefront backward by a certain distance.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront.

        Returns
        -------
        Wavefront
            The wavefront after the propagation.
        '''
        filtered = instance_data.fourier_filter.backward(wavefront.electric_field)

        return Wavefront(filtered, wavefront.wavelength, wavefront.input_stokes_vector)



import numpy as np
from ..util import large_poisson
from ..field import subsample_field, make_supersampled_grid

class Detector(object):
    '''Base class for a detector.

    Parameters
    ----------
    detector_grid : Grid
        The grid on which the detector returns its images. These indicate
        the centers of the pixels.
    subsamping : integer or scalar or ndarray
        The number of subpixels per pixel along one axis. For example, a
        value of 2 indicates that 2x2=4 subpixels are used per pixel. If
        this is a scalar, it will be rounded to the nearest integer. If
        this is an array, the subsampling factor will be different for
        each dimension. Default: 1.

    Attributes
    ----------
    input_grid : Grid
        The grid that is expected as input.
    '''
    def __init__(self, detector_grid, subsamping=1):
        self.detector_grid = detector_grid
        self.subsamping = subsamping

        if subsamping > 1:
            self.input_grid = make_supersampled_grid(detector_grid, subsamping)
        else:
            self.input_grid = detector_grid

    def integrate(self, wavefront, dt, weight=1):
        '''Integrates the detector.

        Parameters
        ----------
        wavefront : Wavefront or array_like
            The wavefront sets the amount of power generated per unit time.
        dt : scalar
            The integration time in units of time.
        weight : scalar
            Weight of every unit of integration time.
        '''
        raise NotImplementedError()

    def read_out(self):
        '''Reads out the detector.

        No noise will be added to the image.

        Returns
        -------
        Field
            The final detector image.
        '''
        raise NotImplementedError()

    def __call__(self, wavefront, dt=1, weight=1):
        '''Integrate and read out the detector.

        This is a convenience function to avoid having to call two functions
        in quick succession.

        Parameters
        ----------
        wavefront : Wavefront or array_like
            The wavefront sets the amount of power generated per unit time.
        dt : scalar
            The integration time in units of time.
        weight : scalar
            Weight of every unit of integration time.

        Returns
        -------
        Field
            The final detector image.
        '''
        self.integrate(wavefront, dt, weight)
        return self.read_out()

class NoiselessDetector(Detector):
    '''A detector without noise.

    This detector includes no noise effects at all. This can be used as a
    theoretically perfect detector.

    Parameters
    ----------
    detector_grid : Grid
        The grid on which the detector returns its images. These indicate
        the centers of the pixels.
    subsamping : integer or scalar or ndarray
        The number of subpixels per pixel along one axis. For example, a
        value of 2 indicates that 2x2=4 subpixels are used per pixel. If
        this is a scalar, it will be rounded to the nearest integer. If
        this is an array, the subsampling factor will be different for
        each dimension. Default: 1.
    '''
    def __init__(self, detector_grid, subsamping=1):
        Detector.__init__(self, detector_grid, subsamping)

        self.accumulated_charge = 0

    def integrate(self, wavefront, dt, weight=1):
        '''Integrates the detector.

        Parameters
        ----------
        wavefront : Wavefront or array_like
            The wavefront sets the amount of power generated per unit time.
        dt : scalar
            The integration time in units of time.
        weight : scalar
            Weight of every unit of integration time.
        '''
        if hasattr(wavefront, 'power'):
            power = wavefront.power
        else:
            power = wavefront

        self.accumulated_charge += power * dt * weight

    def read_out(self):
        '''Reads out the detector.

        No noise will be added to the image.

        Returns
        -------
        Field
            The final detector image.
        '''
        # Make sure not to overwrite output
        output_field = self.accumulated_charge.copy()

        # Reset detector
        self.accumulated_charge = 0

        return output_field

class NoisyDetector(Detector):
    '''A detector class that has some basic noise properties.

    A detector that includes the following noise sources: photon shot noise, dark current
    shot noise, flat field errors and read noise. These noise sources (except for photon shot
    noise) can be either a scalar (same for entire grid) or an array. For the flat field error
    specifically: if a scalar is given, a normal distributed flat field map (given scalar
    is standard deviation about 1) is generated that will be used - if an array is given, that
    array will be used, so it will NOT be a normal distributed flat field where every point in
    array gives the standard deviation. This allows the user to load a specific flat field map.

    The detector can also integrate supersampled wavefronts and subsample them to the correct
    grid size. This allows for the averaging of the power over one detector pixel.

    Parameters
    ----------
    detector_grid : Grid
        The grid on which the detector samples.
    dark_current_rate : scalar or array_like
        An array or scalar that gives dark current rate in counts per unit time for each point
        in the grid.
    read_noise : scalar or array_like
        An array or scalar that gives the rms read noise counts for each point in the grid.
    flat_field : scalar or array_like
        An array or scalar that gives the flat field error for each point in the
        grid. If a scalar is given, a random normal distributed flat field map (given scalar
        is standard deviation about 1) is generated that will be used.
    include_photon_noise : boolean
        Turns the photon noise on or off. Default: True.
    subsampling : integer or scalar or ndarray
        The number of subpixels per pixel along one axis. For example, a
        value of 2 indicates that 2x2=4 subpixels are used per pixel. If
        this is a scalar, it will be rounded to the nearest integer. If
        this is an array, the subsampling factor will be different for
        each dimension. Default: 1.
    '''
    def __init__(self, detector_grid, dark_current_rate=0, read_noise=0, flat_field=0, include_photon_noise=True, subsampling=1):
        Detector.__init__(self, detector_grid, subsampling)

        # Setting the start charge level.
        self.accumulated_charge = 0

        # The parameters.
        self.dark_current_rate = dark_current_rate
        self.read_noise = read_noise
        self.flat_field = flat_field
        self.include_photon_noise = include_photon_noise

    @property
    def flat_field(self):
        return self._flat_field

    @flat_field.setter
    def flat_field(self, flat_field):
        # If the flatfield parameters was a scalar, we will generate a flat field map that will
        # be constant for this object until flat_field is manually changed.
        if np.isscalar(flat_field):
            self._flat_field = np.random.normal(loc=1.0, scale=flat_field, size=self.detector_grid.size)
        else:
            self._flat_field = flat_field

    def integrate(self, wavefront, dt, weight=1):
        '''Integrates the detector.

        The amount of power and dark current that the detector generates are calculated
        for a given integration time and weight.

        Parameters
        ----------
        wavefront : Wavefront or array_like
            The wavefront sets the amount of power generated per unit time.
        dt : scalar
            The integration time in units of time.
        weight : scalar
            Weight of every unit of integration time.
        '''
        # The power that the detector detects during the integration.
        if hasattr(wavefront, 'power'):
            power = wavefront.power
        else:
            power = wavefront

        self.accumulated_charge += subsample_field(power, subsampling=self.subsamping, new_grid=self.detector_grid, statistic='sum') * dt * weight

        # Adding the generated dark current.
        self.accumulated_charge += self.dark_current_rate * dt * weight

    def read_out(self):
        '''Reads out the detector.

        The read out operation of the detector. This means that, if applicable, first photon noise
        will be included to the power. Then, in this order, the flat field error and
        read-out noise are applied. After the read out the power is reset.

        Returns
        -------
        Field
            The final detector image.
        '''
        # Make sure not to overwrite output
        output_field = self.accumulated_charge.copy()

        # Adding photon noise.
        if self.include_photon_noise:
            output_field = large_poisson(output_field, thresh=1e6)

        # Adding flat field errors.
        output_field *= self.flat_field

        # Adding read-out noise.
        output_field += np.random.normal(loc=0, scale=self.read_noise, size=output_field.size)

        # Reset detector
        self.accumulated_charge = 0

        return output_field

class FrameCorrector(object):
    def correct(self, img):
        return img

class BasicFrameCorrector(FrameCorrector):
    def __init__(self, dark=0, flat_field=1):
        self.dark = dark
        self.flat_field = flat_field

    def correct(self, img):
        return (img - self.dark) / self.flat_field



import numpy as np
from ..optics import OpticalElement
from ..mode_basis import ModeBasis
from ..util import inverse_truncated

class PerfectCoronagraph(OpticalElement):
    r'''A perfect coronagraph for a certain aperture and order.

    These type of coronagraphs suppress all light for a flat wavefront. The incoming complex
    amplitude :math:`A` is modified as follows (following [Cavarroc2006]_):

    .. math::
        \overline{A} = A - \Pi \sqrt{S}

    where :math:`\overline{A}` is the resulting complex ampliutude, :math:`\Pi` is the telescope
    pupil, and :math:`S` is the Strehl ratio of the incoming wavefront.

    Higher orders are added by fitting higher-order electric field modes to the incoming
    wavefront and subtracting those, following [Guyon2006]_.

    .. [Cavarroc2006] Celine Cavarroc et al. "Fundamental limitations on Earth-like planet detection with
        extremely large telescopes." Astronomy & Astrophysics 447.1 (2006): 397-403

    .. [Guyon2006] Olivier Guyon et al. "Theoretical limits on extrasolar terrestrial planet detection
        with coronagraphs." The Astrophysical Journal Supplement Series 167.1 (2006): 81

    Parameters
    ----------
    aperture : Field
        The reference aperture. The perfect coronagraph is designed for this aperture.
    order : integer
        The order of the perfect coronagraph. This must be even.
    coeffs : list or ndarray or None
        The coefficients that are used for subtraction. This allows for partial suppression of certain
        modes, which can be used to design perfect coronagraphs that are insensitive to stellar
        radius [Guyon2006]_. If it is None, all modes are completely suppressed.
    '''
    def __init__(self, aperture, order=2, coeffs=None):
        self.pupil_grid = aperture.grid
        modes = []

        if coeffs is not None:
            order = int(2 * np.ceil(0.5 * (np.sqrt(8 * len(coeffs) + 1) - 1)))
            self.coeffs = coeffs
        else:
            self.coeffs = np.ones(int(order * (order / 2 + 1) / 4))

        for i in range(order // 2):
            for j in range(i + 1):
                modes.append(aperture * self.pupil_grid.x**j * self.pupil_grid.y**(i - j))

        self.mode_basis = ModeBasis(modes).orthogonalized

        self.transformation = self.mode_basis.transformation_matrix
        self.transformation_inverse = inverse_truncated(self.transformation, 1e-6)

    def forward(self, wavefront):
        '''Propagate the wavefront through the perfect coronagraph.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront (in the pupil plane)

        Returns
        -------
        Wavefront
            The post-coronagraphic wavefront (in the pupil plane).
        '''
        wf = wavefront.copy()

        # Project the wavefront onto the pupil modes and attenuate by coeffs, then transform back to the pupil.
        # This is done for each polarization separately.
        correction = np.einsum(
            'kj,j,ji,...i->...k',
            self.transformation,
            self.coeffs,
            self.transformation_inverse,
            wf.electric_field,
            optimize='optimal'
        )

        # Then subtract this from the original wavefront to compute the post-coronagraphic field.
        wf.electric_field -= correction

        return wf

    def backward(self, wavefront):
        '''Propagate the wavefront backwards through the perfect coronagraph.

        This method behaves the same as the forward propagation.

        Parameters
        ----------
        wavefront : Wavefront
            The incoming wavefront (in the pupil plane)

        Returns
        -------
        Wavefront
            The post-coronagraphic wavefront (in the pupil plane).
        '''
        return self.forward(wavefront)

    def get_transformation_matrix_forward(self, wavelength=1):
        '''Get the forward propagation transformation matrix.

        Parameters
        ----------
        wavelength : scalar
            The wavelength at which to calculate the transformation matrix.

        Returns
        -------
        ndarray
            The forward transformation_matrix.
        '''
        return np.eye(self.pupil_grid.size) - self.transformation.dot(self.coeffs * self.transformation_inverse)

    def get_transformation_matrix_backward(self, wavelength=1):
        '''Get the backwards propagation transformation matrix.

        This method behaves the same as the forward propagation.

        Parameters
        ----------
        wavelength : scalar
            The wavelength at which to calculate the transformation matrix.

        Returns
        -------
        ndarray
            The backward transformation_matrix.
        '''
        return self.get_transformation_matrix_forward(wavelength)



from .apodization import SurfaceApodizer, PhaseApodizer
import numpy as np
from ..field import Field
from scipy.special import jv

def grating_equation(wavelength, order, period, angle_of_incidence):
    ''' Calculates the angle of the diffracted beam.

    Parameters
    ----------
    wavelength : array like
        The wavelength for which the diffraction efficiency is calculated.
    order : array like
        The diffraction order.
    period : array like
        The period of the diffraction grating.
    angle_of_incidence : array like
        The incidence angles of the incoming field in radians.

    Returns
    -------
    Array like
        The diffraction efficiency
    '''
    return np.arcsin(order * wavelength / period + np.sin(angle_of_incidence))

def diffraction_efficiency_sinusoidal_grating(wavelength, groove_depth, order, period, angle_of_incidence):
    ''' Calculates the diffraction efficiency from a sinusoidal phase grating.

    Parameters
    ----------
    wavelength : array like
        The wavelength for which the diffraction efficiency is calculated.
    groove_depth : arraylike
        The depth of the sinusoidal grating.
    order : array like
        The diffraction order.
    period : array like
        The period of the diffraction grating.
    angle_of_incidence : array like
        The incidence angles of the incoming field in radians.

    Returns
    -------
    Array like
        The diffraction efficiency
    '''
    diffracted_angle = grating_equation(wavelength, order, period, angle_of_incidence)
    phase_difference = np.pi * groove_depth / wavelength * (np.cos(angle_of_incidence) + np.cos(diffracted_angle))
    return jv(order, phase_difference)**2

def snells_law(incidence_angle, relative_refractive_index):
    ''' Applies Snell's law.

    Parameters
    ----------
    incidence_angle : array like
        The incidence angles of the incoming field in radians.
    relative_refractive_index : scalar
        The relative refractive index between two media.

    Returns
    -------
    Array like
        The transmitted angles of the outgoing field in radians.
    '''
    if np.all(relative_refractive_index > 1):
        return np.arcsin(relative_refractive_index * np.sin(incidence_angle))
    else:
        if np.all(incidence_angle < np.arcsin(relative_refractive_index)):
            return np.arcsin(relative_refractive_index * np.sin(incidence_angle))
        else:
            raise ValueError("Total internal reflection is occuring.")

class TiltElement(SurfaceApodizer):
    ''' An element that applies a tilt.

    Parameters
    ----------
    angle: scalar
        The tilt angle in radians.
    orientation : scalar
        The orientation of the tilt in radians. The default orientation is aligned along the y-axis.
    refractive_index : scalar or function
        The refractive index of the material. The default is 2.0 which makes it achromatic and exact.
    '''
    def __init__(self, angle, orientation=0, refractive_index=2.0):
        self._angle = angle
        self._orientation = orientation
        super().__init__(self.tilt_sag, refractive_index)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, new_angle):
        self._angle = new_angle
        self.surface_sag = self.tilt_sag

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, new_orientation):
        self._orientation = new_orientation
        self.surface_sag = self.tilt_sag

    def tilt_sag(self, grid):
        ''' Calculate the sag profile for the tilt element.

        Parameters
        ----------
        grid : Grid
            The grid on which the surface sag is calculated.

        Returns
        -------
        Field
            The surface sag.
        '''
        return Field(grid.rotated(self._orientation).y * np.tan(self._angle), grid)

class ThinPrism(TiltElement):
    '''A thin prism that operates in the paraxial regime.

    Parameters
    ----------
    angle : scalar
        The wedge angle of the prism in radians.
    orientation : scalar
        The orientation of the prism in radians. The default orientation is aligned along the x-axis.
    refractive_index : scalar or function of wavelength
        The refractive index of the prism.
    '''
    def __init__(self, angle, refractive_index, orientation=0):
        super().__init__(angle, orientation, refractive_index)

    def minimal_deviation_angle(self, wavelength):
        ''' Find the angle of minimal deviation for a paraxial prism.

        Parameters
        ----------
        wavelength : scalar
            The wavelength that is traced through the prism.

        Returns
        -------
        scalar
            The angle of minimal deviation in radians.
        '''
        n = self._refractive_index(wavelength)
        return (n - 1) * self._prism_angle

    def trace(self, wavelength):
        ''' Trace a paraxial ray through the prism.

        Parameters
        ----------
        wavelength : scalar
            The wavelength that is traced through the prism.

        Returns
        -------
        scalar
            The angle of deviation for the traced ray in radians.
        '''
        return (self._refractive_index(wavelength) - 1) * self.angle

class Prism(SurfaceApodizer):
    '''A prism that deviates the beam.

    Parameters
    ----------
    angle_of_incidence : sacalar
        The angle of incidence of the wavefront in radians.
    prism_angle : scalar
        The angle of the prism in radians.
    refractive_index : scalar or function of wavelength
        The refractive index of the prism.
    orientation : scalar
        The orientation of the prism in radians. The default orientation is aligned along the y-axis.
    '''
    def __init__(self, angle_of_incidence, prism_angle, refractive_index, orientation=0):
        self._prism_angle = prism_angle
        self._angle_of_incidence = angle_of_incidence
        self._refractive_index = refractive_index
        self._orientation = orientation

        super().__init__(self.prism_sag, self._refractive_index)

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, new_orientation):
        self._orientation = new_orientation
        self.surface_sag = self.prism_sag

    @property
    def prism_angle(self):
        return self._prism_angle

    @prism_angle.setter
    def prism_angle(self, new_prism_angle):
        self._prism_angle = new_prism_angle
        self.surface_sag = self.prism_sag

    def minimal_deviation_angle(self, wavelength):
        ''' Find the angle of minimal deviation for a prism.

        Parameters
        ----------
        wavelength : scalar
            The wavelength that is traced through the prism.

        Returns
        -------
        scalar
            The angle of minimal deviation in radians.
        '''
        n = self._refractive_index(wavelength)
        return 2 * np.arcsin(n * np.sin(self._prism_angle / 2)) - self._prism_angle

    def trace(self, wavelength):
        ''' Trace a ray through the prism.

        Parameters
        ----------
        wavelength : scalar
            The wavelength that is traced through the prism.

        Returns
        -------
        scalar
            The angle of deviation for the traced ray in radians.
        '''
        n = self._refractive_index(wavelength)
        transmitted_angle_surface_1 = snells_law(self._angle_of_incidence, 1 / n)

        incident_angle_surface_2 = self._prism_angle - transmitted_angle_surface_1
        transmitted_angle = snells_law(incident_angle_surface_2, n)

        angle_of_deviation = self._angle_of_incidence + transmitted_angle - self._prism_angle

        return angle_of_deviation

    def prism_sag(self, grid, wavelength):
        ''' Calculate the sag profile for the prism.

        Parameters
        ----------
        grid : Grid
            The grid on which the surface sag is calculated.
        wavelength : scalar
            The wavelength for which the surface sag is calculated.

        Returns
        -------
        Field
            The surface sag.
        '''
        theta = self.trace(wavelength)
        return Field(grid.rotated(self._orientation).y * np.tan(theta) / (self._refractive_index(wavelength) - 1), grid)

class PhaseGrating(PhaseApodizer):
    '''A grating that applies an achromatic phase pattern.

    Parameters
    ----------
    grating_period : scalar
        The period of the grating.
    grating_amplitude : scalar
        The amplitude of the grating.
    grating_profile : field generator
        The profile of the grating. The default is None and assumes a sinusoidal profile for the grating.
    orientation : scalar
        The orientation of the grating in radians. The default orientation is aligned along the y-axis.
    '''
    def __init__(self, grating_period, grating_amplitude, grating_profile=None, orientation=0):
        self._grating_period = grating_period
        self._orientation = orientation
        self._grating_amplitude = grating_amplitude

        if grating_profile is None:
            def sinusoidal_grating_profile(grid):
                return np.sin(2 * np.pi * grid.y)

            grating_profile = sinusoidal_grating_profile

        self._grating_profile = grating_profile

        super().__init__(self.grating_pattern)

    def grating_pattern(self, grid):
        return self._grating_amplitude * Field(self._grating_profile(grid.rotated(self._orientation).scaled(1 / self._grating_period)), grid)

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, new_orientation):
        self._orientation = new_orientation
        self.phase = self.grating_pattern

    @property
    def period(self):
        return self._grating_period

    @period.setter
    def period(self, new_period):
        self._grating_period = new_period
        self.phase = self.grating_pattern

    @property
    def amplitude(self):
        return self._amplitude

    @amplitude.setter
    def amplitude(self, new_amplitude):
        self._amplitude = new_amplitude
        self.phase = self.grating_pattern



import numpy as np

def get_strehl_from_focal(img, ref_img):
    '''Get the Strehl ratio from a focal-plane image.

    Parameters
    ----------
    img : Field or array_like
        The focal-plane image.
    ref_img : Field or array_like
        The reference focal-plane image without aberrations.

    Returns
    -------
    scalar
        The Strehl ratio.
    '''
    return img[np.argmax(ref_img)] / ref_img.max()

def get_strehl_from_pupil(aperture, ref_aperture):
    '''Get the Strehl ratio from a pupil-plane electric field.

    Parameters
    ----------
    aperture : Field or array_like
        The pupil-plane electric field.
    ref_aperture : Field or array_like
        The reference pupil-plane electric field without aberrations.

    Returns
    -------
    scalar
        The Strehl ratio.
    '''
    return np.abs(np.sum(aperture) / np.sum(ref_aperture))**2

def get_mean_intensity_in_roi(img, mask):
    '''Get the mean intensity in a masked region of interest.

    Parameters
    ----------
    img : Field or array_like
        The focal-plane image.
    mask : Field or array_like
        A binary array describing the region of interest.

    Returns
    -------
    scalar
        The mean intensity in the region of interest.
    '''
    return np.mean(img[mask])

def get_mean_raw_contrast(img, mask, ref_img):
    '''Get the mean raw contrast in a masked region of interest.

    Parameters
    ----------
    img : Field or array_like
        The focal-plane image.
    mask : Field or array_like
        A binary array describing the region of interest.
    img_ref : Field or array_like
        A reference focal-plane image without aberrations. This is used
        to determine the Strehl ratio.

    Returns
    -------
    scalar
        The mean raw contrast in the region of interest.
    '''
    mean_intensity = get_mean_intensity_in_roi(img, mask)
    strehl = get_strehl_from_focal(img, ref_img)

    return mean_intensity / strehl



import numpy as np
from scipy.fft import next_fast_len

from .fourier_transform import _get_float_and_complex_dtype
from .._math import fft as _fft_module

class ChirpZTransform:
    '''The Chirp Z-transform (CZT).

    This class evaluates:

    .. math:: X_k = \sum_{n=0}^{N-1} x(n) z_k^{-n}

    where :math:`N` is the number of points along the real space, :math:`M`
    is the number of points in the Z-plane. The Chirp Z-transform samples
    the Z-plane on equidistant points along spiral arcs, where the arcs are
    given by the starting point :math:`A` and the complex ratio between points
    :math:`W`:

    .. math:: z_k = A * W^{-k}, k=0, 1, \ldots, M - 1

    This implementation uses Bluestein's algorithm to compute the CZT using Fast
    Fourier Transforms.

    Parameters
    ----------
    n : integer
        The number of points in the real plane.
    m : integer
        The number of points in the Z-plane.
    w : scalar
        The complex ratio between points in the Z-plane. Note: ensure that its
        absolute value is close to one, so that w**(max(n, m)**2) does not
        overflow.
    a : scalar
        The starting point in the complex Z-plane. Note: ensure that its absolute
        value is close to one, so that a**max(n, m) does not overflow.
    '''
    def __init__(self, n, m, w, a):
        # Ensure that w and a are complex scalars.
        w = complex(w)
        a = complex(a)

        self.n = n
        self.m = m
        self.w = w
        self.a = a

        self._current_dtype = None

    def _compute_kernel_and_weights(self, dtype):
        _, complex_dtype = _get_float_and_complex_dtype(dtype)

        if complex_dtype != self._current_dtype:
            k = np.arange(max(self.m, self.n))
            self.k = k

            wk2 = self.w**(self.k**2 / 2)

            self.nfft = next_fast_len(self.n + self.m - 1)

            self._Awk2 = self.a**-k[:self.n] * wk2[:self.n]
            self._Fwk2 = _fft_module.fft(1 / np.hstack((wk2[self.n - 1:0:-1], wk2[:self.m])), self.nfft)
            self._wk2 = wk2[:self.m]
            self._yidx = slice(self.n - 1, self.n + self.m - 1)

            self._Awk2 = self._Awk2.astype(complex_dtype, copy=False)
            self._Fwk2 = self._Fwk2.astype(complex_dtype, copy=False)
            self._wk2 = self._wk2.astype(complex_dtype, copy=False)

            self._current_dtype = complex_dtype

    def __call__(self, x):
        '''Compute the Chirp Z-transform along the last axis.

        The input is expected to have `n` points along the last axis.
        This is assumed but not checked. The transform is multiplexed
        along all other axes.

        The return value has the same number of axes, with the last axis
        having `m` elements.

        Parameters
        ----------
        x : array_like
            The array to compute the chirp Z-transform for.

        Returns
        -------
        array_like
            The chirp Z-transformed array.
        '''
        self._compute_kernel_and_weights(x.dtype)

        # Perform the CZT.
        x = x * self._Awk2

        intermediate = _fft_module.fft(x, self.nfft)
        intermediate *= self._Fwk2
        res = _fft_module.ifft(intermediate)

        res = res[..., self._yidx] * self._wk2

        _, complex_dtype = _get_float_and_complex_dtype(x.dtype)
        return res.astype(complex_dtype, copy=False)



from .optical_element import AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward

import numpy as np

class Magnifier(AgnosticOpticalElement):
    def __init__(self, magnification):
        ''' An ideal magnifier.

        This class magnifies a wavefront in an energy conserving manner.

        Parameters
        ----------
        magnification : scalar or function of wavelength
            The magnification of the system.
        '''
        self._magnification = magnification

        AgnosticOpticalElement.__init__(self, False, True)

    def make_instance(self, instance_data, input_grid, output_grid, wavelength):
        instance_data.magnification = self.evaluate_parameter(self._magnification, input_grid, output_grid, wavelength) * np.ones(2)

    @property
    def magnification(self):
        return self._magnification

    @magnification.setter
    def magnification(self, magnification):
        self._magnification = magnification

        self.clear_cache()

    def get_input_grid(self, output_grid, wavelength):
        magnification = self.evaluate_parameter(self._magnification, None, None, wavelength) * np.ones(2)

        return output_grid.scaled(1.0 / magnification)

    def get_output_grid(self, input_grid, wavelength):
        magnification = self.evaluate_parameter(self._magnification, input_grid, None, wavelength) * np.ones(2)

        return input_grid.scaled(magnification)

    @make_agnostic_forward
    def forward(self, instance_data, wavefront):
        wf = wavefront.copy()

        wf.electric_field.grid = wf.electric_field.grid.scaled(instance_data.magnification)
        wf.electric_field /= np.sqrt(np.prod(instance_data.magnification))

        return wf

    @make_agnostic_backward
    def backward(self, instance_data, wavefront):
        wf = wavefront.copy()

        wf.electric_field.grid = wf.electric_field.grid.scaled(1.0 / instance_data.magnification)
        wf.electric_field *= np.sqrt(np.prod(instance_data.magnification))

        return wf



import numpy as np
from .grid import Grid
from .coordinates import UnstructuredCoords

from functools import reduce
import operator

def _prod(iterable):
    return reduce(operator.mul, iterable, 1)

def _get_rotation_matrix(ndim, angle, axis=None):
    if ndim == 1:
        raise ValueError('Rotation of a one-dimensional grid is not possible.')
    elif ndim > 3:
        raise NotImplementedError()

    if ndim == 2:
        return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
    elif ndim == 3:
        if axis is None:
            raise ValueError('An axis must be supplied when rotating a three-dimensional grid.')

        axis = np.array(axis).astype('float')
        if np.all(np.array(axis.shape) != (3,)):
            raise ValueError('The axis must be a 3-vector.')
        axis /= np.sqrt(axis.dot(axis))

        K = np.array([[0, -axis[2], axis[1]], [axis[2], 0, -axis[0]], [-axis[1], axis[0], 0]])
        return np.eye(3) + np.sin(angle) * K + (1 - np.cos(angle)) * K.dot(K)

class CartesianGrid(Grid):
    '''A grid representing a N-dimensional Cartesian coordinate system.
    '''
    _coordinate_system = 'cartesian'

    @property
    def x(self):
        '''The x-coordinate (dimension 0).
        '''
        return self.coords[0]

    @property
    def y(self):
        '''The y-coordinate (dimension 1).
        '''
        return self.coords[1]

    @property
    def z(self):
        '''The z-coordinate (dimension 2).
        '''
        return self.coords[2]

    @property
    def w(self):
        '''The w-coordinate (dimension 3).
        '''
        return self.coords[3]

    def scale(self, scale):
        '''Scale the grid in-place.

        Parameters
        ----------
        scale : array_like or scalar
            The factor with which to scale the grid.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        if np.isscalar(scale):
            self.weights *= np.abs(scale)**self.ndim
        else:
            self.weights *= np.prod(np.abs(scale))

        self.coords *= scale

        return self

    def shift(self, shift):
        '''Shift the grid in-place.

        Parameters
        ----------
        shift : array_like
            The amount with which to shift the grid.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        self.coords += shift
        return self

    def rotate(self, angle, axis=None):
        '''Rotate the grid in-place.

        .. caution::
            All structure in the coordinates will be destroyed.

        Parameters
        ----------
        angle : scalar
            The angle in radians.
        axis : ndarray or None
            The axis of rotation. For two-dimensional grids, it is ignored. For
            three-dimensional grids it is required.

        Returns
        -------
        Grid
            Itself to allow for chaining these transformations.
        '''
        R = _get_rotation_matrix(self.ndim, angle, axis)

        coords = np.einsum('ik,kn->in', R, np.array(self.coords))
        self.coords = UnstructuredCoords(coords)
        return self

    def rotated(self, angle, axis=None):
        '''A rotated copy of this grid.

        Parameters
        ----------
        angle : scalar
            The angle in radians.
        axis : ndarray or None
            The axis of rotation. For two-dimensional grids, it is ignored. For
            three-dimensional grids it is required.

        Returns
        -------
        Grid
            The rotated grid.
        '''
        R = _get_rotation_matrix(self.ndim, angle, axis)

        coords = np.einsum('ik,kn->in', R, np.array(self.coords))
        return CartesianGrid(UnstructuredCoords(coords))

    @staticmethod
    def _get_automatic_weights(coords):
        if coords.is_regular:
            return np.prod(coords.delta)
        elif coords.is_separated:
            weights = []
            for i in range(len(coords)):
                x = coords.separated_coords[i]
                w = (x[2:] - x[:-2]) / 2.
                w = np.concatenate(([x[1] - x[0]], w, [x[-1] - x[-2]]))
                weights.append(w)

            return _prod(np.ix_(*weights[::-1])).ravel()

Grid._add_coordinate_system('cartesian', CartesianGrid)



import numpy as np
import scipy.sparse

from ..field import Field, Grid

class ModeBasis(object):
    '''A linear basis of modes.

    Modes can be stored both in a dense or sparse format. A sparse mode basis reduces memory usage when modes
    are localized and contain a lot of zeros. Projection onto and linear combinations of modes are done seamlessly,
    regardless of the sparsity of the mode basis. If a grid is available, linear combinations automatically become
    Fields.

    You can create your own mode bases by supplying a list of modes, or a transformation matrix that transforms from
    mode coefficients to the mode. You can also use a number of built-in mode bases in HCIPy.

    Parameters
    ----------
    transformation_matrix : array_like or list of array_like
        The transformation matrix of the mode basis or a list of modes.
    grid : Grid or None
        The grid on which the modes are defined.
    '''
    def __init__(self, transformation_matrix, grid=None):
        if scipy.sparse.issparse(transformation_matrix):
            sparse = True
            is_list = False
        elif scipy.sparse.issparse(transformation_matrix[0]):
            sparse = np.all([transformation_matrix[i].shape[0] == 1 for i in range(len(transformation_matrix))])
            is_list = True
        else:
            sparse = False
            is_list = isinstance(transformation_matrix, (list, tuple))

        if sparse:
            if is_list:
                self._modes = scipy.sparse.vstack(transformation_matrix, format='csr')
                self._transformation_matrix = self._modes.T.tocsc()
            else:
                self._modes = transformation_matrix.T.tocsr()
                self._transformation_matrix = transformation_matrix.tocsc()
        else:
            if is_list:
                self._transformation_matrix = np.asarray(np.stack(transformation_matrix, axis=-1))
            else:
                self._transformation_matrix = np.asarray(transformation_matrix)

        if grid is not None:
            self.grid = grid
        elif hasattr(transformation_matrix[0], 'grid'):
            self.grid = transformation_matrix[0].grid
        else:
            self.grid = None

    @classmethod
    def from_dict(cls, tree):
        '''Make a ModeBasis from a dictionary, previously created by `to_dict()`.

        Parameters
        ----------
        tree : dictionary
            The dictionary from which to make a new ModeBasis object.

        Returns
        -------
        ModeBasis
            The created object.

        Raises
        ------
        ValueError
            If the dictionary is not formatted correctly.
        '''
        if isinstance(tree['transformation_matrix'], dict):
            data = np.array(tree['transformation_matrix']['data'])
            indices = np.array(tree['transformation_matrix']['indices'])
            indptr = np.array(tree['transformation_matrix']['indptr'])
            shape = tree['transformation_matrix']['shape']

            transformation_matrix = scipy.sparse.csc_matrix((data, indices, indptr), shape=shape)
        else:
            transformation_matrix = np.array(tree['transformation_matrix'])

        if 'grid' in tree:
            mode_basis = cls(transformation_matrix, Grid.from_dict(tree['grid']))
        else:
            mode_basis = cls(transformation_matrix)

        if tree['is_sparse']:
            return mode_basis.to_sparse()
        else:
            return mode_basis.to_dense()

    def to_dict(self):
        '''Convert the object to a dictionary for serialization.

        Returns
        -------
        dictionary
            The created dictionary.
        '''
        if self.is_sparse:
            transformation_matrix = {
                'data': self._transformation_matrix.data,
                'indices': self._transformation_matrix.indices,
                'indptr': self._transformation_matrix.indptr,
                'shape': list(self._transformation_matrix.shape)
            }
        else:
            transformation_matrix = self.transformation_matrix

        tree = {
            'grid': self.grid.to_dict(),
            'transformation_matrix': transformation_matrix,
            'is_sparse': self.is_sparse
        }

        return tree

    @property
    def is_sparse(self):
        '''If the mode basis is sparse.
        '''
        return scipy.sparse.issparse(self._transformation_matrix)

    @property
    def is_dense(self):
        '''If the mode basis is dense.
        '''
        return not self.is_sparse

    def to_sparse(self, copy=False):
        '''Convert the mode basis to a sparse mode basis.

        Parameters
        ----------
        copy : boolean
            Whether to force a copy or not. A copy is always made if
            the current ModeBasis is not sparse.

        Returns
        -------
        ModeBasis
            The sparsified ModeBasis.

        Raises
        ------
            TypeError
                If this ModeBasis cannot be sparsified.
        '''
        if self.is_sparse:
            if copy:
                return ModeBasis(self._transformation_matrix.copy(), self.grid)
            else:
                return self
        else:
            if self._transformation_matrix.ndim != 2:
                raise TypeError('Cannot sparsify a mode basis of tensor fields')

            T = scipy.sparse.csc_matrix(self._transformation_matrix)
            T.eliminate_zeros()
            return ModeBasis(T, self.grid)

    def to_dense(self, copy=False):
        '''Convert the mode basis to a dense mode basis.

        Parameters
        ----------
        copy : boolean
            Whether to force a copy or not. A copy is always made if
            the current ModeBasis is not dense.

        Returns
        -------
        ModeBasis
            The densified ModeBasis.
        '''
        if self.is_dense:
            if copy:
                return ModeBasis(self._transformation_matrix.copy(), self.grid)
            else:
                return self
        else:
            T = np.asarray(self._transformation_matrix.todense())
            return ModeBasis(T, self.grid)

    @property
    def transformation_matrix(self):
        '''The transformation matrix of this mode basis.
        '''
        return self._transformation_matrix

    @transformation_matrix.setter
    def transformation_matrix(self, transformation_matrix):
        self._transformation_matrix = transformation_matrix

    def coefficients_for(self, b, dampening_factor=0):
        r'''Calculate the coefficients on this mode basis in a least squares fashion.

        The vector `b` is projection onto the mode basis in a least squares fashion. This
        means that the const function

        .. math:: J(c) = |b - A x|^2_2 + |\lambda x|^2_2

        is minimized, where :math:`x` are the coefficients, and :math:`\lambda` is the
        dampening factor.

        If this projection needs to be done repeatedly, you may be better off calculating
        the inverse of the transformation matrix directly and left-multiplying that with
        your vector, rather than using a least squares estimation every time.

        Parameters
        ----------
        b : array_like or Field
            The vector for which to calculate the coefficients.
        dampening_factor : scalar
            The Tikhonov dampening factor used for the least squares procedure.

        Returns
        -------
        array_like
            The coefficients that correspond to the vector `b`.
        '''
        if self.is_sparse or dampening_factor != 0:
            x, istop, itn, normr, normar, norma, conda, normx = scipy.sparse.linalg.lsmr(self._transformation_matrix, b, damp=dampening_factor)
            return x
        else:
            x, residuals, rank, s = np.linalg.lstsq(self._transformation_matrix, b)
            return x

    def linear_combination(self, coefficients):
        '''Calculate a linear combination using this mode basis.

        Parameters
        ----------
        coefficients : array_like or list
            The coefficients of the linear combinations.

        Returns
        -------
        array_like or Field
            The calculated linear combination.
        '''
        y = self._transformation_matrix.dot(coefficients)

        if self.grid is None:
            return y
        else:
            return Field(y, self.grid)

    @property
    def orthogonalized(self):
        '''Get an orthogonalized version of this ModeBasis.

        The resulting ModeBasis spans the same vector space, but each mode is orthogonal to
        all others. In general the resulting `ModeBasis` is dense, so no distinction is made
        between sparse and dense mode bases in this function. This function will always return
        a dense mode basis.

        Returns
        -------
        ModeBasis
            A mode basis with orthogonalized modes.

        Raises
        ------
        NotImplementedError
            If the mode basis is a mode basis containing non-scalar fields.
        '''
        if self._transformation_matrix.ndim != 2:
            raise NotImplementedError('The mode basis contains non-scalar fields; orthogonalization is not implemented for these.')

        q, r = np.linalg.qr(self._transformation_matrix)
        return ModeBasis(q, self.grid)

    def __getitem__(self, item):
        '''Get the `item`-th mode in the `ModeBasis`.

        Parameters
        ----------
        item : int or slice or array_like
            The index/indices of the mode(s).

        Returns
        -------
        Field or array_like or ModeBasis
            The `item`-th mode in the `ModeBasis`.
        '''
        T = self._transformation_matrix[..., item]

        return_mode_basis = False
        if self.is_sparse:
            if T.shape[-1] != 1:
                return_mode_basis = True
        if self.is_dense:
            if T.ndim == self._transformation_matrix.ndim:
                return_mode_basis = True

        if return_mode_basis:
            # We are returning multiple modes; put these in a ModeBasis.
            return ModeBasis(T, self.grid)
        else:
            # We are returning a single mode; return just this.
            if self.is_sparse:
                T = T.toarray()[..., 0]

            if self.grid is None:
                return T
            else:
                return Field(T, self.grid)

    def __len__(self):
        '''The number of modes in the `ModeBasis`.

        Returns
        -------
        int
            The number of modes in the `ModeBasis`.
        '''
        return self.transformation_matrix.shape[-1]

    @property
    def num_modes(self):
        '''The number of modes in the `ModeBasis`.
        '''
        return len(self)

    def append(self, mode):
        '''Append `mode` to this mode basis.

        Parameters
        ----------
        mode : array_like or Field
            The mode to add to the ModeBasis
        '''
        if self.is_sparse:
            self._transformation_matrix = scipy.sparse.hstack((self._transformation_matrix, mode), 'csc')
        else:
            self._transformation_matrix = np.concatenate((self._transformation_matrix, [mode]), axis=-1)

    def extend(self, modes):
        '''Extend the mode basis with `modes`.

        Parameters
        ----------
        modes : list or array_like or ModeBasis
            The modes to add to the ModeBasis.
        '''
        if isinstance(modes, ModeBasis):
            modes = modes.transformation_matrix

        if self.is_sparse:
            self._transformation_matrix = scipy.sparse.hstack((self._transformation_matrix, modes), 'csc')
        else:
            # TODO: worry about modes being in a list instead of transformation matrix or mode basis
            self._transformation_matrix = np.concatenate((self._transformation_matrix, modes), axis=-1)

    def __add__(self, mode_basis):
        '''Merge two mode bases into one.

        Parameters
        ----------
        mode_basis : ModeBasis
            The ModeBasis to add.

        Returns
        -------
        ModeBasis
            The newly created `ModeBasis`.
        '''
        if self.is_sparse or mode_basis.is_sparse:
            transformation_matrix = scipy.sparse.hstack((self._transformation_matrix, mode_basis.transformation_matrix), 'csc')
        else:
            transformation_matrix = np.concatenate((self._transformation_matrix, mode_basis.transformation_matrix), axis=-1)
        return ModeBasis(transformation_matrix, self.grid)



import copy
import numpy as np
import numexpr as ne

from ..field import Field, field_dot, field_kron

_U_matrix = 1 / np.sqrt(2) * np.array([
    [1, 0, 0, 1],
    [1, 0, 0, -1],
    [0, 1, 1, 0],
    [0, 1j, -1j, 0]])

# TODO Should add a pilot Gaussian beam with each Wavefront

class Wavefront(object):
    '''A physical wavefront in an optical system.

    This represents the state of light to be propagated through the
    optical system. It can be both an electric field in the scalar
    approximation (ie. scalar wavefront propgation), a fully polarized
    wavefront, represented by a Field of Jones vectors, and a potentially
    partially-polarized wavefront, represented by two Jones vector fields
    and the Stokes vector corresponding to the Jones vectors (1,0) and (0,1).

    Parameters
    ----------
    electric_field : Field
        The electric field, either scalar, vectorial or 2-tensorial.
    wavelength : scalar
        The wavelength of the wavefront.
    input_stokes_vector : ndarray or None
        If a Stokes vector (I, Q, U, V) is given, a partially-polarized
        wavefront is initialized. If `electric_field` is scalar, it will
        be transformed into a tensor field with the correct Jones states.
        If a tensor-field is given as the `electric_field`, the electric
        field will be interpreted as the Jones matrix modifying the input
        Stokes vector.

    Raises
    ------
    ValueError
        When a Stokes vector is supplied but a vector field is given as
        electric field, or when an input Stokes vector is not supplied,
        but a 2-tensor field is given as electric field.
    '''
    def __init__(self, electric_field, wavelength=1, input_stokes_vector=None):
        if input_stokes_vector is not None:
            if electric_field.tensor_order not in [0, 2]:
                raise ValueError('When supplying a Stokes vector, the electric field must be either a scalar or 2-tensor field.')

            if electric_field.is_scalar_field:
                self.electric_field = electric_field[np.newaxis, np.newaxis, :] * np.eye(2)[..., np.newaxis]
            else:
                self.electric_field = electric_field

            self._input_stokes_vector = np.array(input_stokes_vector)
        else:
            self.electric_field = electric_field
            self._input_stokes_vector = None

            if electric_field.tensor_order == 2:
                raise ValueError('When supplying a 2-tensor field as electric field, an input Stokes vector is required.')

        self.wavelength = wavelength

    def copy(self):
        '''Make a copy of the wavefront.
        '''
        return copy.deepcopy(self)

    @property
    def electric_field(self):
        '''The electric field as function of 2D position on the plane.
        '''
        return self._electric_field

    @electric_field.setter
    def electric_field(self, electric_field):
        if not hasattr(electric_field, 'grid'):
            raise ValueError('The electric field must be a Field.')

        # Cast to complex with correct bit depth
        if electric_field.dtype == 'float32' or electric_field.dtype == 'complex64':
            dtype = 'complex64'
        else:
            dtype = 'complex128'

        self._electric_field = electric_field.astype(dtype, copy=False)

    @property
    def input_stokes_vector(self):
        '''The Stokes vector corresponding to the Jones vectors (1,0) and (0,1).
        '''
        return self._input_stokes_vector

    @property
    def wavenumber(self):
        '''The wavenumber of the light.
        '''
        return 2 * np.pi / self.wavelength

    @wavenumber.setter
    def wavenumber(self, wavenumber):
        self.wavelength = 2 * np.pi / wavenumber

    @property
    def grid(self):
        '''The grid on which the electric field is defined.
        '''
        return self.electric_field.grid

    @property
    def I(self):  # noqa: N802
        '''The I-component of the Stokes vector as function of 2D position
        in the plane.
        '''
        if self.is_scalar:
            # This is a scaler field.
            intensity = ne.evaluate('real(abs(elec))**2', local_dict={'elec': self.electric_field})
            return Field(intensity, self.electric_field.grid)
        elif self.is_partially_polarized:
            # This is a tensor field.
            x = self._electric_field[0, 0, :]
            y = self._electric_field[0, 1, :]
            z = self._electric_field[1, 0, :]
            w = self._electric_field[1, 1, :]

            a, b, c, d = self._input_stokes_vector

            # NumExpr is not smart enough to have abs(x) be real, so we have to take the real
            # part manually. This may change in later releases of NumExpr.
            M11 = 'real(abs(x))**2 + real(abs(y))**2 + real(abs(z))**2 + real(abs(w))**2'
            M12 = 'real(abs(x))**2 - real(abs(y))**2 + real(abs(z))**2 - real(abs(w))**2'
            M13 = '2 * (real(x) * real(y) + imag(x) * imag(y) + real(z) * real(w) + imag(z) * imag(w))'
            M14 = '2 * (-real(x) * imag(y) + imag(x) * real(y) - real(z) * imag(w) + imag(z) * real(w))'

            res = '0.5 * ((' + M11 + ') * a + (' + M12 + ') * b + (' + M13 + ') * c +  (' + M14 + ') * d)'
            local_dict = {'x': x, 'y': y, 'z': z, 'w': w, 'a': a, 'b': b, 'c': c, 'd': d}

            return Field(ne.evaluate(res, local_dict=local_dict), self.electric_field.grid)
        else:
            # This is a vector field.
            return np.sum(np.abs(self.electric_field)**2, axis=0)

    @property
    def Q(self):  # noqa: N802
        '''The Q-component of the Stokes vector as function of 2D position
        in the plane.
        '''
        if self.is_scalar:
            # This is a scaler field.
            return self.grid.zeros()
        elif self.is_partially_polarized:
            # This is a tensor field.
            x = self._electric_field[0, 0, :]
            y = self._electric_field[0, 1, :]
            z = self._electric_field[1, 0, :]
            w = self._electric_field[1, 1, :]

            a, b, c, d = self._input_stokes_vector

            M21 = 'real(abs(x))**2 + real(abs(y))**2 - real(abs(z))**2 - real(abs(w))**2'
            M22 = 'real(abs(x))**2 - real(abs(y))**2 - real(abs(z))**2 + real(abs(w))**2'
            M23 = '2 * (real(x) * real(y) + imag(x) * imag(y) - real(z) * real(w) - imag(z) * imag(w))'
            M24 = '2 * (-real(x) * imag(y) + imag(x) * real(y) + real(z) * imag(w) - imag(z) * real(w))'

            res = '0.5 * ((' + M21 + ') * a + (' + M22 + ') * b + (' + M23 + ') * c +  (' + M24 + ') * d)'
            local_dict = {'x': x, 'y': y, 'z': z, 'w': w, 'a': a, 'b': b, 'c': c, 'd': d}

            return Field(ne.evaluate(res, local_dict=local_dict), self.electric_field.grid)
        else:
            # This is a vector field.
            return np.abs(self.electric_field[0, :])**2 - np.abs(self.electric_field[1, :])**2

    @property
    def U(self):  # noqa: N802
        '''The U-component of the Stokes vector as function of 2D position
        in the plane.
        '''
        if self.is_scalar:
            # This is a scaler field.
            return self.grid.zeros()
        elif self.is_partially_polarized:
            # This is a tensor field.
            x = self._electric_field[0, 0, :]
            y = self._electric_field[0, 1, :]
            z = self._electric_field[1, 0, :]
            w = self._electric_field[1, 1, :]

            a, b, c, d = self._input_stokes_vector

            M31 = '2 * (real(x) * real(z) + imag(x) * imag(z) + real(y) * real(w) + imag(y) * imag(w))'
            M32 = '2 * (real(x) * real(z) + imag(x) * imag(z) - real(y) * real(w) - imag(y) * imag(w))'
            M33 = '2 * (real(x) * real(w) + imag(x) * imag(w) + real(y) * real(z) + imag(y) * imag(z))'
            M34 = '2 * (imag(x) * real(w) - real(x) * imag(w) + real(y) * imag(z) - imag(y) * real(z))'

            res = '0.5 * ((' + M31 + ') * a + (' + M32 + ') * b + (' + M33 + ') * c +  (' + M34 + ') * d)'
            local_dict = {'x': x, 'y': y, 'z': z, 'w': w, 'a': a, 'b': b, 'c': c, 'd': d}

            return Field(ne.evaluate(res, local_dict=local_dict), self.electric_field.grid)
        else:
            # This is a vector field.
            return 2 * np.real(self.electric_field[0, :] * self.electric_field[1, :].conj())

    @property
    def V(self):  # noqa: N802
        '''The V-component of the Stokes vector as function of 2D position
        in the plane.
        '''
        if self.is_scalar:
            # This is a scaler field.
            return self.grid.zeros()
        elif self.is_partially_polarized:
            # This is a tensor field.
            x = self._electric_field[0, 0, :]
            y = self._electric_field[0, 1, :]
            z = self._electric_field[1, 0, :]
            w = self._electric_field[1, 1, :]

            a, b, c, d = self._input_stokes_vector

            M41 = '2 * (real(x) * imag(z) - imag(x) * real(z) + real(y) * imag(w) - imag(y) * real(w))'
            M42 = '2 * (real(x) * imag(z) - imag(x) * real(z) - real(y) * imag(w) + imag(y) * real(w))'
            M43 = '2 * (real(x) * imag(w) - imag(x) * real(w) + real(y) * imag(z) - imag(y) * real(z))'
            M44 = '2 * (real(x) * real(w) + imag(x) * imag(w) - real(y) * real(z) - imag(y) * imag(z))'

            res = '0.5 * ((' + M41 + ') * a + (' + M42 + ') * b + (' + M43 + ') * c +  (' + M44 + ') * d)'
            local_dict = {'x': x, 'y': y, 'z': z, 'w': w, 'a': a, 'b': b, 'c': c, 'd': d}
            return Field(ne.evaluate(res, local_dict=local_dict), self.electric_field.grid)
        else:
            # This is a vector field.
            return -2 * np.imag(self.electric_field[0, :] * self.electric_field[1, :].conj())

    @property
    def stokes_vector(self):
        '''The Stokes vector.
        '''
        if self.is_scalar:
            # This is a scalar field and thus we return an unpolarized Stokes vector.
            stokes_vector = Field(np.zeros((4, self.grid.size)), self.grid)
            stokes_vector[0, :] = np.abs(self.electric_field)**2

            return stokes_vector
        elif self.is_partially_polarized:
            # This is a tensor field.
            mueller_matrix = jones_to_mueller(self.electric_field)

            return field_dot(mueller_matrix, self._input_stokes_vector)
        else:
            # This is a vector field and thus we return a fully polarized Stokes vector.
            stokes_vector = Field(np.zeros((4, self.grid.size)), self.grid)

            stokes_vector[0, :] = self.I
            stokes_vector[1, :] = self.Q
            stokes_vector[2, :] = self.U
            stokes_vector[3, :] = self.V

            return stokes_vector

    @property
    def degree_of_polarization(self):
        '''The degree of polarization.
        '''
        return np.sqrt(self.Q**2 + self.U**2 + self.V**2) / self.I

    @property
    def degree_of_linear_polarization(self):
        '''The degree of linear polarization.
        '''
        return np.sqrt(self.Q**2 + self.U**2) / self.I

    @property
    def angle_of_linear_polarization(self):
        '''The angle of linear polarization.
        '''
        return 0.5 * np.arctan2(self.U, self.Q)

    @property
    def degree_of_circular_polarization(self):
        '''The degree of circular polarization.
        '''
        return self.V / self.I

    @property
    def ellipticity(self):
        '''The ratio of the minor to major axis of the electric
        field polarization ellipse.
        '''
        return self.V / (self.I + np.sqrt(self.Q**2 + self.U**2))

    @property
    def is_polarized(self):
        '''If the wavefront can be polarized.
        '''
        return self.electric_field.tensor_order in [1, 2]

    @property
    def is_partially_polarized(self):
        '''If the wavefront can be partially polarized.
        '''
        return self.electric_field.tensor_order == 2

    @property
    def is_scalar(self):
        '''If the wavefront uses the scalar approximation.
        '''
        return self.electric_field.is_scalar_field

    @property
    def intensity(self):
        '''The total intensity of the wavefront as function of 2D position on the plane.
        '''
        return self.I

    @property
    def amplitude(self):
        '''The amplitude of the wavefront as function of 2D position on the plane.
        '''
        return np.abs(self.electric_field)

    @property
    def phase(self):
        '''The phase of the wavefront as function of 2D position on the plane.
        '''
        phase = np.angle(self.electric_field)
        return Field(phase, self.electric_field.grid)

    @property
    def real(self):
        '''The real part of the wavefront as function of 2D position on the plane.
        '''
        return np.real(self.electric_field)

    @property
    def imag(self):
        '''The imaginary part of the wavefront as function of 2D position on the plane.
        '''
        return np.imag(self.electric_field)

    @property
    def power(self):
        '''The power of each pixel in the wavefront.
        '''
        if self.electric_field.is_scalar_field or self.electric_field.is_vector_field:
            variables = {'field': self.electric_field, 'weights': self.grid.weights}
            power = ne.evaluate('real(abs(field))**2 * weights', local_dict=variables)

            return Field(power, self.grid)
        else:
            return self.intensity * self.grid.weights

    @property
    def total_power(self):
        '''The total power in this wavefront.
        '''
        return np.sum(self.power)

    @total_power.setter
    def total_power(self, p):
        self.electric_field *= np.sqrt(p / self.total_power)

def jones_to_mueller(jones_matrix):
    '''Convert a Jones matrix to a Mueller matrix.

    Parameters
    ----------
    jones_matrix : ndarray or tensor field
        The Jones matrix/matrices to convert to Mueller matrix/matrices.

    Returns
    -------
    ndarray or tensor Field
        The Mueller matrix/matrices.
    '''
    return np.real(field_dot(_U_matrix, field_dot(field_kron(jones_matrix, jones_matrix.conj()), _U_matrix.conj().T)))



from cycler import cycler
from . import colors

# Blue, red, green, orange, purple, brown
palettes = {
    'dark': cycler(color=[colors.blue_700, colors.red_700, colors.green_700, colors.orange_700, colors.purple_700, colors.brown_700])
}

def set_color_scheme(dark=False, publication_quality=False, cmap='viridis'):  # pragma: no cover
    """
    Apply a color scheme to all matplotlib figures. The setting
    publication_quality uses LaTeX for all text in the figure.
    """
    import warnings
    warnings.warn(
        'set_color_scheme() is deprecated. Copy the color scheme to your own file if you want to continue using it in the future.',
        DeprecationWarning,
        stacklevel=2
    )

    import matplotlib as mpl

    mpl.rc('lines', linewidth=1.5, markeredgewidth=0.25)
    mpl.rc('image', cmap=cmap)
    mpl.rc('legend', scatterpoints=1, numpoints=1, labelspacing=0.3)
    mpl.rc('axes.formatter', limits=(-4, 4))
    mpl.rc('text.latex', preamble=['\\usepackage{amsmath}'])

    mpl.rc('xtick', labelsize='small')
    mpl.rc('ytick', labelsize='small')
    mpl.rc('axes', titlesize='medium', labelsize='medium')
    mpl.rc('legend', fontsize='medium')

    mpl.rc('savefig', transparent=True)

    if dark:
        mpl.rc('axes', prop_cycle=palettes['light'], facecolor='k', labelcolor='w', edgecolor='w')
        mpl.rc('xtick', color='w')
        mpl.rc('ytick', color='w')
        mpl.rc('grid', color='w')
        mpl.rc('figure', facecolor='k', edgecolor='k')
        mpl.rc('text', color='w')
    else:
        mpl.rc('axes', prop_cycle=palettes['dark'], facecolor='w', labelcolor='k', edgecolor='k')
        mpl.rc('xtick', color='k')
        mpl.rc('ytick', color='k')
        mpl.rc('grid', color='k')
        mpl.rc('figure', facecolor='w', edgecolor='w')
        mpl.rc('text', color='k')

    if publication_quality:
        mpl.rc('text', usetex=True)
        mpl.rc('font', family='sans-serif')
        mpl.rc('font', serif=['computer modern roman'], monospace=['computer modern typewriter'])
        mpl.rcParams['font.sans-serif'] = ['computer modern sans serif']

        mpl.rc('font', size=11)
        mpl.rc('figure', figsize=(7.2, 5.1))
    else:
        mpl.rc('text', usetex=False)
        mpl.rc('font', family='sans-serif')
        mpl.rc('font', serif=[
            'Bitstream Vera Serif', 'New Century Schoolbook', 'Century Schoolbook L', 'Utopia',
            'ITC Bookman', 'Bookman', 'Nimbus Roman No9 L', 'Times New Roman', 'Times', 'Palatino', 'Charter', 'serif'])
        mpl.rcParams['font.sans-serif'] = [
            'Bitstream Vera Sans', 'Lucida Grande', 'Verdana', 'Geneva', 'Lucid', 'Arial',
            'Helvetica', 'Avant Garde', 'sans-serif']
        mpl.rc('font', monospace=[
            'Bitstream Vera Sans Mono', 'Andale Mono', 'Nimbus Mono L', 'Courier New', 'Courier',
            'Fixed', 'Terminal', 'monospace'])

        mpl.rc('figure', figsize=(10, 7.1))
        mpl.rc('font', size=14)



import numpy as np
from scipy.special import jv, jn_zeros, jnp_zeros

from .mode_basis import ModeBasis
from ..aperture import make_circular_aperture
from ..field import Field

def disk_harmonic(n, m, D=1, bc='dirichlet', grid=None):
    '''Create a disk harmonic.

    Parameters
    ----------
    n : int
        Radial order
    m : int
        Azimuthal order
    D : scalar
        The diameter of the pupil.
    bc : string
        The boundary conditions to use. This can be either 'dirichlet', or
        'neumann' for a Dirichlet or Neumann boundary condition respectively.
    grid : Grid
        The grid on which to evaluate the function.

    Returns
    -------
    Field
        The disk harmonic function evaluated on `grid`.

    Raises
    ------
    ValueError
        If the boundary condition is not recognized.
    '''
    polar_grid = grid.as_('polar')
    r = 2 * polar_grid.r / D
    theta = polar_grid.theta

    m_negative = m < 0
    m = abs(m)

    if bc == 'dirichlet':
        lambda_mn = jn_zeros(m, n)[-1]
        norm = 1
    elif bc == 'neumann':
        lambda_mn = jnp_zeros(m, n)[-1]
        norm = 1
    else:
        raise ValueError('Boundary condition not recognized.')

    if m_negative:
        z = norm * jv(m, lambda_mn * r) * np.sin(m * theta)
    else:
        z = norm * jv(m, lambda_mn * r) * np.cos(m * theta)

    # Do manual normalization for now...
    mask = make_circular_aperture(D)(grid) > 0.5
    norm = np.sqrt(np.sum(z[mask]**2))

    return Field(z * mask / norm, grid)

def disk_harmonic_energy(n, m, bc='dirichlet'):
    '''Get the energy of a disk harmonic function.

    This allows for functions to sort a disk harmonic mode basis on energy.

    Parameters
    ----------
    n : int
        Radial order
    m : int
        Azimuthal order
    bc : string
        The boundary conditions to use. This can be either 'dirichlet', or
        'neumann' for a Dirichlet or Neumann boundary condition respectively.

    Returns
    -------
    scalar
        The energy corresponding to the mode.

    Raises
    ------
    ValueError
        If the boundary condition is not recognized.
    '''
    m = abs(m)

    if bc == 'dirichlet':
        lambda_mn = jn_zeros(m, n)[-1]
    elif bc == 'neumann':
        lambda_mn = jnp_zeros(m, n)[-1]
    else:
        raise ValueError('Boundary condition not recognized.')

    return lambda_mn**2

def get_disk_harmonic_orders_sorted(num_modes, bc='dirichlet'):
    '''Return the orders for the first `num_modes` modes.

    Parameters
    ----------
    num_modes : int
        The number of modes to return.
    bc : string
        The boundary conditions to use. This can be either 'dirichlet', or
        'neumann' for a Dirichlet or Neumann boundary condition respectively.

    Returns
    -------
    list of tuples
        A list of orders (n, m).
    '''
    orders = [(1, 0)]
    energies = [disk_harmonic_energy(1, 0, bc)]

    results = []
    while len(results) < num_modes:
        k = np.argmin(energies)
        order = orders[k]

        if order[1] != 0:
            results.append((order[0], -order[1]))
        results.append(order)

        del orders[k]
        del energies[k]

        new_order = (order[0], order[1] + 1)
        if new_order not in results and new_order not in orders:
            orders.append(new_order)
            energies.append(disk_harmonic_energy(new_order[0], new_order[1], bc))
        new_order = (order[0] + 1, order[1])
        if new_order not in results and new_order not in orders:
            orders.append(new_order)
            energies.append(disk_harmonic_energy(new_order[0], new_order[1], bc))

    return results[:num_modes]

def make_disk_harmonic_basis(grid, num_modes, D=1, bc='dirichlet'):
    '''Create a disk harmonic mode basis.

    Parameters
    ----------
    grid : Grid
        The grid on which to evaluate the disk harmonic modes.
    num_modes : int
        The number of modes to create.
    D : scalar
        The diameter of the disk.
    bc : string
        The boundary conditions to use. This can be either 'dirichlet', or
        'neumann' for a Dirichlet or Neumann boundary condition respectively.

    Returns
    -------
    ModeBasis
        The evaluated disk harmonic modes.
    '''
    orders = get_disk_harmonic_orders_sorted(num_modes, bc)

    modes = [disk_harmonic(order[0], order[1], D, bc, grid) for order in orders]
    return ModeBasis(modes, grid)
